/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#define BOOST_CONTROLLER_PWM_BITS 8
#define BOOST_CONTROLLER_INTERP_SHIFT 18
#define BOOST_CONTROLLER_DUTY_BITS (BOOST_CONTROLLER_PWM_BITS + BOOST_CONTROLLER_INTERP_SHIFT)
#define BOOST_CONTROLLER_PWM_COUNT (1 << BOOST_CONTROLLER_PWM_BITS)
#define BOOST_CONTROLLER_DUTY_MAX (1 << BOOST_CONTROLLER_DUTY_BITS)
#define BOOST_CONTROLLER_DUTY_LIMIT (95 * (BOOST_CONTROLLER_DUTY_MAX / 100))
#define BOOST_CONTROLLER_RATE_LIMIT 96  // 3V/ms with 31.25kHz update rate
#define BOOST_CONTROLLER_PID_P_GAIN_OVERVOLT 10000 // The P coeff if over top V

#include <stdbool.h>
#include <stdint.h>

#include "boostController.h"

static uint32_t target;
static uint32_t maxOutput;
static uint16_t Pnom;
static uint16_t Inom;
static volatile uint32_t internalTarget;
static volatile int32_t accumulator;
static volatile int32_t dutyCycle;
static volatile int32_t ditherError;
static volatile uint32_t pidPTerm;
static volatile uint32_t pidITerm;

static void setPIDTerms();

void boostController_init(uint32_t initialOutputmv,
                          uint32_t maxOutputmv,
                          uint16_t pidPNom,
                          uint16_t pidINom)
{
    internalTarget = initialOutputmv;
    target = initialOutputmv;
    Pnom = pidPNom;
    Inom = pidINom;

    maxOutput = maxOutputmv;

    accumulator = 0;
    dutyCycle = 0;
    ditherError = 0;
}

void boostController_setTarget(uint32_t targetmv)
{
    target = targetmv;
}

void boostController_getNextSamples(uint32_t actualmv, 
                                    uint32_t* sampleBuffer,
                                    uint8_t sampleCount)
{
    setPIDTerms();

    // Respond quickly if we exceed the maximum output
    if(actualmv > maxOutput) accumulator = 0;

    // Update internal target
    int32_t targetDiff = target - internalTarget;
    targetDiff = targetDiff > BOOST_CONTROLLER_RATE_LIMIT ?
                              BOOST_CONTROLLER_RATE_LIMIT : targetDiff;
    targetDiff = targetDiff < (-1 * BOOST_CONTROLLER_RATE_LIMIT) ?
                              (-1 * BOOST_CONTROLLER_RATE_LIMIT) : targetDiff;

    internalTarget += targetDiff;

    // Calculate the error, and limit
    int32_t error = internalTarget - actualmv;

    //
    // Calculate P
    //
    dutyCycle = error * pidPTerm; 

    //
    // Calculate I
    //
    accumulator += error * pidITerm;
    accumulator = accumulator < 0 ? 0 : accumulator;
    accumulator = accumulator > BOOST_CONTROLLER_DUTY_LIMIT ? 
                                BOOST_CONTROLLER_DUTY_LIMIT : accumulator;
    dutyCycle += accumulator;

    //
    // Clamp the PID output value to usable duty cycles
    //
    dutyCycle = dutyCycle < 0 ? 0 : dutyCycle;

    dutyCycle = dutyCycle > BOOST_CONTROLLER_DUTY_LIMIT ? 
                            BOOST_CONTROLLER_DUTY_LIMIT : dutyCycle;

    // Calculate actual timer counter compare values, using dither
    // to improve resolution
    for(uint8_t idx=0; idx<sampleCount; idx++)
    {
        sampleBuffer[idx] = (dutyCycle + ditherError) >> 
                            BOOST_CONTROLLER_INTERP_SHIFT;

        ditherError += dutyCycle - (sampleBuffer[idx] << 
                                    BOOST_CONTROLLER_INTERP_SHIFT);
    }
}

static void setPIDTerms()
{
    static const uint16_t calcShift = 12;
    static const uint16_t calcSize = 1 << 12;

    if(dutyCycle < (BOOST_CONTROLLER_DUTY_MAX/2))
    {
        pidPTerm = Pnom / 2;
        pidITerm = Inom / 2;
        return;
    }

    pidPTerm = Pnom * (calcSize - (dutyCycle >> (BOOST_CONTROLLER_DUTY_BITS 
                                                 - calcShift)));
    pidPTerm = pidPTerm >> calcShift;

    pidITerm = Inom * (calcSize - (dutyCycle >> (BOOST_CONTROLLER_DUTY_BITS 
                                                 - calcShift)));
    pidITerm = pidITerm >> calcShift;

}

