/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#include <stdlib.h>
#include <math.h>

#include "discPumpCore.h"

typedef struct discPumpCore
{
    discPumpCore_setup setup;

    //Power target
    uint32_t targetPowerMicrowatts;
    uint16_t maxPowerMilliwatts;
    bool dutyCyclePowerControl;
    bool driveEnabled;

    //The current target drive in millivolts
    int32_t driveMillivolts;

    //The current drive frequency
    uint16_t freqHz;

    //Frequency sweep variables
    bool sweepPerformed;
    bool sweepStarted;
    uint8_t preSweepSettleTicks;
    uint16_t bestFreqHz;
    uint32_t bestTrackingValue;

    //Frequency Tracking Variables
    bool frequencyTrackingEnabled;
    uint16_t manualDriveFrequency;
    bool frequencyTrackingPreviouslyEnabled;
    uint32_t prevTrackingValue;
    uint16_t trackDirectionChangeHz;
    int8_t trackDirection;

    //Current system status
    discPumpCore_status status;
} discPumpCore;

static int32_t clamp(int32_t min, int32_t max, int32_t val);

static void powerTracking(discPumpCore_handle handle,
                          discPumpCore_update update);

static void powerTrackingVoltage(discPumpCore_handle handle,
                                 uint16_t measuredDriveMillivolts,
                                 uint32_t measuredDriveMicroamps);

static void powerTrackingSoftDrive(discPumpCore_handle handle,
                                   discPumpCore_update update);

static void powerTrackingDutyCycle(discPumpCore_handle handle,
                                   uint16_t measuredDriveMillivolts,
                                   uint32_t measuredDriveMicroamps,
                                   uint16_t actualDutyCycle);

static void getNewFrequency(discPumpCore_handle handle, 
                            discPumpCore_update update);
                                
static void sweepRunner(discPumpCore_handle handle, 
                        discPumpCore_update update);

static float getSoftDriveFeature(discPumpCore_update update);

discPumpCore_handle discPumpCore_init(discPumpCore_setup setup)
{
    discPumpCore_handle handle = malloc(sizeof(discPumpCore));

    handle->setup = setup;
    handle->freqHz = 20000; 
    handle->sweepPerformed = false;
    handle->sweepStarted = false;
    handle->preSweepSettleTicks = 10;

    handle->frequencyTrackingPreviouslyEnabled = true;
    handle->frequencyTrackingEnabled = true;
    handle->manualDriveFrequency = 21000;
    handle->prevTrackingValue = 1000000;
    handle->trackDirectionChangeHz = 0;
    handle->trackDirection = setup.driveFreqTrackingStepSizeHz;

    handle->driveMillivolts = 0;

    handle->maxPowerMilliwatts = 1000;
    handle->dutyCyclePowerControl = true;

    handle->status = DISC_PUMP_CORE_STATUS_OKAY;

    return handle;
}

void discPumpCore_destroy(discPumpCore_handle handle)
{
    free(handle);
}

discPumpCore_status discPumpCore_setPower(discPumpCore_handle handle,
                                          uint32_t targetPowerMicrowatts)
{
    handle->targetPowerMicrowatts = clamp(0, 
                                    handle->maxPowerMilliwatts * 1000, 
                                    targetPowerMicrowatts);
    handle->driveEnabled = true;

    return DISC_PUMP_CORE_STATUS_OKAY;
}

discPumpCore_status discPumpCore_setPowerLimit(discPumpCore_handle handle,
                                               uint16_t maxPowerMilliwatts)
{
    handle->maxPowerMilliwatts = maxPowerMilliwatts;

    return handle->status;
}

discPumpCore_status discPumpCore_setManualFrequency(discPumpCore_handle handle,
                                                    uint16_t freqHz)
{
    handle->frequencyTrackingEnabled = false;
    handle->manualDriveFrequency = freqHz;

    return handle->status;
}

discPumpCore_status discPumpCore_enableFrequencyTracking(discPumpCore_handle 
                                                                        handle)
{
    handle->frequencyTrackingEnabled = true;
    return handle->status;
}

discPumpCore_status discPumpCore_disable(discPumpCore_handle handle)
{
    if(handle->status != DISC_PUMP_CORE_STATUS_OKAY)
    {
        handle->sweepPerformed = false;
        handle->status = DISC_PUMP_CORE_STATUS_OKAY;
    }

    handle->setup.disableDrive();
    handle->driveEnabled = false;
    handle->driveMillivolts = handle->setup.minDriveMillivolts;

    return DISC_PUMP_CORE_STATUS_OKAY;
}

discPumpCore_status discPumpCore_tick(discPumpCore_handle handle,
                                      discPumpCore_update update)
{
    if(!handle->driveEnabled) return DISC_PUMP_CORE_STATUS_OKAY;


    //Trigger a sweep if we've just trasitioned from manual frequency
    //to frequency tracking
    if(!handle->frequencyTrackingPreviouslyEnabled &&
       handle->frequencyTrackingEnabled)
    {
        handle->sweepPerformed = false;
    }
    //update previous state of freq tracking
    handle->frequencyTrackingPreviouslyEnabled = handle->frequencyTrackingEnabled;

    //Perform frequency sweep if one hasn't been performed
    if(handle->sweepPerformed == false)
    {
        sweepRunner(handle, update);
        return DISC_PUMP_CORE_STATUS_OKAY;
    }

    //Raise error if overcurrent
    if(update.measuredDriveMicroamps > 135000 ||
       handle->status == DISC_PUMP_CORE_STATUS_ERR_SHORT)
    {
        handle->status = DISC_PUMP_CORE_STATUS_ERR_SHORT;
        handle->setup.disableDrive();

        return DISC_PUMP_CORE_STATUS_ERR_SHORT;
    }

    //Perform frequency tracking if appropriate
    if(update.measuredDriveMicroamps > 5000 && 
       handle->frequencyTrackingEnabled)
    {
        getNewFrequency(handle, update);
    } 
    else if(!handle->frequencyTrackingEnabled)
    {
        handle->freqHz = handle->manualDriveFrequency;
    }

    //Check the target frequencies
    if(handle->freqHz > handle->setup.maxDriveFreq ||
       handle->status == DISC_PUMP_CORE_STATUS_ERR_OVER_FREQ)
    {
        handle->status = DISC_PUMP_CORE_STATUS_ERR_OVER_FREQ;
        handle->setup.disableDrive();

        return DISC_PUMP_CORE_STATUS_ERR_OVER_FREQ;
    }

    if(handle->freqHz < handle->setup.minDriveFreq ||
       handle->status == DISC_PUMP_CORE_STATUS_ERR_UNDER_FREQ)
    {
        handle->status = DISC_PUMP_CORE_STATUS_ERR_UNDER_FREQ;
        handle->setup.disableDrive();

        return DISC_PUMP_CORE_STATUS_ERR_UNDER_FREQ;
    }

    //Perform power tracking
    powerTracking(handle, update);

    return DISC_PUMP_CORE_STATUS_OKAY;
} 

//TODO - see if we can avoid floating point math in this function
static void powerTracking(discPumpCore_handle handle,
                          discPumpCore_update update)
{

    uint16_t measuredDriveMillivolts = update.measuredDriveMillivolts;
    uint32_t measuredDriveMicroamps = update.measuredDriveMicroamps;
    uint16_t actualDutyCycle = update.actualDutyCycle;

    //Just set a low target if measured values are very low,
    //when not in duty cycle mode
    if(!handle->dutyCyclePowerControl && measuredDriveMillivolts < 200)
    {
        handle->setup.setDrive(handle->freqHz, 
                               handle->setup.minDriveMillivolts, 
                               0xFFFF);
        return;
    }

    //Perform a different form of power tracking based upon whether
    //we're controlling with voltage or duty cycle
    if(handle->setup.useSoftDriveProtocol)
    {
        powerTrackingSoftDrive(handle, update);
        return;
    }

    if(handle->dutyCyclePowerControl)
    {
        powerTrackingDutyCycle(handle, 
                               measuredDriveMillivolts,
                               measuredDriveMicroamps,
                               actualDutyCycle);
        return;
    }

    powerTrackingVoltage(handle,
                         measuredDriveMillivolts,
                         measuredDriveMicroamps);
}

static void powerTrackingVoltage(discPumpCore_handle handle,
                                 uint16_t measuredDriveMillivolts,
                                 uint32_t measuredDriveMicroamps)
{
    float impedance = (measuredDriveMillivolts * 1000.0) /
                      measuredDriveMicroamps;

    int32_t targetmv = sqrt(handle->targetPowerMicrowatts * impedance);

    int32_t voltageErr = targetmv - measuredDriveMillivolts;
    int32_t voltageErrAbs = voltageErr >= 0 ?  // Get magnitude of the error
                            voltageErr : voltageErr * -1;

    // Higher gain for smaller errors
    if(voltageErrAbs < 500)
        voltageErr *= 0.8;
    else if(voltageErrAbs < 1000)
        voltageErr *= 0.7;
    else if(voltageErrAbs < 1500)
        voltageErr *= 0.6;
    else if(voltageErrAbs < 2000)
        voltageErr *= 0.5;
    else if(voltageErrAbs < 3000)
        voltageErr *= 0.4;
    else if(voltageErrAbs < 5000)
        voltageErr *= 0.35;
    else
        voltageErr *= 0.3;

    //Clamp voltage error to sensible range
    int32_t maxRampRate = handle->setup.maxDriveRampRateMillivolts;
    voltageErr = clamp(-8000, 8000, voltageErr);
    voltageErr = clamp(maxRampRate * -1, maxRampRate, voltageErr);

    if(handle->setup.driveVoltageOpenLoop)
    {
        handle->driveMillivolts = voltageErr + measuredDriveMillivolts;
    }
    else
    {
        handle->driveMillivolts += voltageErr;
    }

    //Limit drive voltage
    handle->driveMillivolts = clamp(handle->setup.minDriveMillivolts, 
                                    handle->setup.maxDriveMillivolts,
                                    handle->driveMillivolts);

    if(handle->driveMillivolts == handle->setup.minDriveMillivolts)
    {
        handle->dutyCyclePowerControl = true;
    }

    handle->setup.setDrive(handle->freqHz, handle->driveMillivolts, 0xFFFF);
}

static void powerTrackingSoftDrive(discPumpCore_handle handle,
                                   discPumpCore_update update)
{
    float realCurrentua = update.measuredDriveMicroamps * 
                          cosf(update.phaseMilliDegrees * 
                               (M_PI * 2.0 / 360000));
        
    float impedance = (update.measuredDriveMillivoltsACrms * 1000.0) /
                      realCurrentua;

    int32_t targetmv = sqrt(handle->targetPowerMicrowatts * impedance);

    int32_t voltageErr = targetmv - update.measuredDriveMillivoltsACrms;
    
    // Shift by half the error
    // TODO - revert to 0.5
    voltageErr *= 0.1;

    // Clamp voltage error so we don't exceed the ACrms limit
    voltageErr = clamp(-9999, handle->setup.maxDriveMillivoltsACrms - 
                              update.measuredDriveMillivoltsACrms,
                       voltageErr);

    // Clamp voltage error to sensible range
    voltageErr = clamp(-3000, 3000, voltageErr);
    
    // Clamp voltage to user specified max ramp rate
    int32_t maxRampRate = handle->setup.maxDriveRampRateMillivolts;
    voltageErr = clamp(maxRampRate * -1, maxRampRate, voltageErr);
        
    // Apply change in the boost voltage
    handle->driveMillivolts += voltageErr;

    // Limit drive voltage
    handle->driveMillivolts = clamp(handle->setup.minDriveMillivolts, 
                                    handle->setup.maxDriveMillivolts,
                                    handle->driveMillivolts);

    handle->setup.setDrive(handle->freqHz, handle->driveMillivolts, 0xFFFF);
}

static void powerTrackingDutyCycle(discPumpCore_handle handle,
                                   uint16_t measuredDriveMillivolts,
                                   uint32_t measuredDriveMicroamps,
                                   uint16_t actualDutyCycle)
{
    float poweruW = (float)measuredDriveMillivolts * 
                    (float)measuredDriveMicroamps / 1000.0;
    
    int32_t duty = actualDutyCycle + 
                   (handle->targetPowerMicrowatts - poweruW) / 10;

    duty = clamp(handle->setup.minDutyCycle,
                 0xFFFF,
                 duty);

    if(duty == 0xFFFF)
    {
        handle->dutyCyclePowerControl = false;
    }

    handle->setup.setDrive(handle->freqHz,
                           handle->setup.minDriveMillivolts,
                           duty);
}

static void getNewFrequency(discPumpCore_handle handle, 
                            discPumpCore_update update)
{
    if(handle->trackDirectionChangeHz >=
       handle->setup.driveFreqTrackingHysteresisHz)
    {
        uint32_t trackingValue;

        if(handle->setup.useSoftDriveProtocol)
        {
            trackingValue = 1000000 / getSoftDriveFeature(update);
        }
        else
        {
            trackingValue = (50000 * update.measuredDriveMillivolts) / 
                            (update.measuredDriveMicroamps / 20);
        }

        if(handle->prevTrackingValue < trackingValue)
        {
            handle->trackDirection *= -1;
            handle->trackDirectionChangeHz = 0;
        }

        handle->prevTrackingValue = trackingValue;
    }

    if(handle->trackDirectionChangeHz < 1000) //avoid overflow
    {
        handle->trackDirectionChangeHz += 
            handle->setup.driveFreqTrackingStepSizeHz;
    }

    handle->freqHz += handle->trackDirection;
}

static void sweepRunner(discPumpCore_handle handle, 
                        discPumpCore_update update)
{
    uint16_t volts = update.measuredDriveMillivolts;
    uint32_t currentMicroamps = update.measuredDriveMicroamps;

    if(handle->sweepStarted == false)
    {
        handle->freqHz = handle->setup.minDriveFreq;
        handle->bestTrackingValue = 0;
        handle->setup.setDrive(handle->freqHz, 
                               handle->setup.minDriveMillivolts, 
                               0xFFFF);

        //Wait till voltage within 1000mV of min drive voltage
        //Then wait for the pre sweep settle ticks
        if(volts > handle->setup.minDriveMillivolts - 1000 && 
           volts < handle->setup.minDriveMillivolts + 1000)
        {
            handle->preSweepSettleTicks--;
        }

        if(handle->preSweepSettleTicks <= 0)
            handle->sweepStarted = true;

        return;
    }

    uint32_t trackingFeature = handle->setup.useSoftDriveProtocol ? 
                               getSoftDriveFeature(update) : currentMicroamps;

    if(trackingFeature > handle->bestTrackingValue)
    {
        handle->bestTrackingValue = trackingFeature;
        handle->bestFreqHz = handle->freqHz;
    }

    handle->freqHz += 50;

    if(handle->freqHz >= handle->setup.maxDriveFreq)
    {
        handle->freqHz = handle->bestFreqHz;
        handle->sweepPerformed = true;
        handle->sweepStarted = false;
        handle->preSweepSettleTicks = 25;

        return;
    }

    handle->setup.setDrive(handle->freqHz, 
                           handle->setup.minDriveMillivolts, 
                           0xFFFF);
}

static float getSoftDriveFeature(discPumpCore_update update)
{
    // Get current in quadrature form
    float uA_real = update.measuredDriveMicroamps * 
                    cosf(update.phaseMilliDegrees * 
                         (M_PI * 2.0 / 360000));
    float uA_img = update.measuredDriveMicroamps * 
                   sinf(update.phaseMilliDegrees * 
                        (M_PI * 2.0 / 360000));

    // Calculate expected C0 current
    float omega = update.actualDriveFreqHz * 2.0 * M_PI;
    float uA_C0 = update.measuredDriveMillivoltsACrms * omega * 5.5E-9 * 1E3;

    // Subtract to get iLCR
    uA_img -= uA_C0;
    float iLCR = sqrt(pow(uA_img, 2) + pow(uA_real, 2));

    // Calculate real power
    float pwr_real = uA_real * update.measuredDriveMillivoltsACrms; 

    // return sqrt(real power) / iLCR
    return 10000 * sqrt(pwr_real) / iLCR;
}

static int32_t clamp(int32_t min, int32_t max, int32_t val)
{
    if(val > max)
    {
        return max;
    }
    else if(val < min)
    {
        return min;
    }

    return val;
}

