/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#ifndef _DISCPUMPCORE_H
#define _DISCPUMPCORE_H

#include <stdbool.h>
#include <stdint.h>

// Enum giving the status of the driven disc pump
typedef enum 
{
    DISC_PUMP_CORE_STATUS_OKAY,
    DISC_PUMP_CORE_STATUS_ERR_SHORT,
    DISC_PUMP_CORE_STATUS_ERR_OVER_FREQ,
    DISC_PUMP_CORE_STATUS_ERR_UNDER_FREQ
} discPumpCore_status;

// A struct used to pass setup details to the discPumpCore driver
typedef struct discPumpCore_setup
{
    // Callbacks to allow discPumpCore to set the output to the pump
    void (*setDrive)(uint16_t frequencyHz,
                     uint32_t voltageMillivolts,
                     uint16_t dutyCycle); //Out of 0xFFFF

    void (*disableDrive)(void);

    // Settings to limit the voltage from the boost converter, ramp rate of the
    // boost converter, and the minimum duty cycle that can be achieved
    uint32_t minDriveMillivolts;
    uint32_t maxDriveMillivolts;
    uint16_t minDutyCycle; //Out of 0xFFFF
    uint16_t maxDriveRampRateMillivolts;
    bool driveVoltageOpenLoop;

    // Settings for using the "Soft Drive" protocol, that requires the hardware
    // to measure the phase between the driving voltage and current.
    // Often the voltage accross the pump using typical "Soft Drive" hardware
    // is not the same as the output of the boost converter. Therefore you can
    // set a limit on the max allowable RMS voltage over the pump
    bool useSoftDriveProtocol;
    uint32_t maxDriveMillivoltsACrms;  // Only required for soft drive

    // The minimum and maximum frequencies the pump can be driven at.
    // An error is raised if the algorithms track the pumps frequency to
    // one of these limits
    uint16_t minDriveFreq;
    uint16_t maxDriveFreq;

    // Step size, and minimum distance to move in one direction for the
    // random walk algorithm used to track the drive frequency
    uint8_t driveFreqTrackingHysteresisHz;  //Suggest ~15Hz
    uint8_t driveFreqTrackingStepSizeHz;    //Suggest ~2Hz
} discPumpCore_setup;

typedef struct discPumpCore_update
{
    // Values required for normal Disc Pump operation
    uint32_t measuredDriveMillivolts;
    uint32_t measuredDriveMicroamps;
    uint16_t actualDriveFreqHz;
    uint16_t actualDutyCycle; //Out of 0xFFFF

    // Further values required if running "Soft Drive" protocol
    int32_t phaseMilliDegrees;
    uint32_t measuredDriveMillivoltsACrms;
} discPumpCore_update;

// A handle to reference a specific disc pump
// This allows several pumps to be driven individually
typedef struct discPumpCore* discPumpCore_handle;


// Setup a new discPumpCore driver, and get a handle for the pump
discPumpCore_handle discPumpCore_init(discPumpCore_setup setup);

// Frees the memory used by the discPumpCore driver referenced by the handle
void discPumpCore_destroy(discPumpCore_handle handle);

// Set a target power for a pump
discPumpCore_status discPumpCore_setPower(discPumpCore_handle handle,
                                          uint32_t targetPowerMicrowatts);

// Set a limit on the target power, useful if you have a control loop calling
// discPumpCore_setPower, and you want to limit the maximum power it can set
discPumpCore_status discPumpCore_setPowerLimit(discPumpCore_handle handle,
                                               uint16_t maxPowerMilliwatts);

// Forces the pump to drive at a fixed frequency
discPumpCore_status discPumpCore_setManualFrequency(discPumpCore_handle handle,
                                                    uint16_t freqHz);

// Enables frequency tracking (default option)
discPumpCore_status discPumpCore_enableFrequencyTracking(discPumpCore_handle 
                                                                        handle);

// Disable the disc pump drive
discPumpCore_status discPumpCore_disable(discPumpCore_handle handle);

// Tick the discPumpCore driver
// This should be called at a regular interval
// Ideally at 1kHz
discPumpCore_status discPumpCore_tick(discPumpCore_handle handle,
                                      discPumpCore_update update);

#endif // _DISCPUMPCORE_H
