/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */


#include "GPIO_Handler.h"


// helper values for timer counter states
static uint32_t TC_PWMMode1;
static uint32_t TC_PWMMode2;
static uint32_t TC_ForceActive;
static uint32_t TC_ForceInactive;


// Called once to give the GPIO Handler the hardware states for Timer Counter, as they are the same for all TC
void GPIO_Handler_set_TC_states(uint32_t set_TC_PWMMode1, uint32_t set_TC_PWMMode2, uint32_t set_TC_ForceActive, uint32_t set_TC_ForceInactive)
{
	// compiler hack to write to a const variable
	/*uint32_t* temp1 = (uint32_t*)(&TC_PWMMode1);
	(*temp1) = set_TC_PWMMode1;
	uint32_t* temp2 = (uint32_t*)(&TC_PWMMode2);
	(*temp2) = set_TC_PWMMode2;
	uint32_t* temp3 = (uint32_t*)(&TC_ForceActive);
	(*temp3) = set_TC_ForceActive;
	uint32_t* temp4 = (uint32_t*)(&TC_ForceInactive);
	(*temp4) = set_TC_ForceInactive;*/
	TC_PWMMode1 = set_TC_PWMMode1;
	TC_PWMMode2 = set_TC_PWMMode2;
	TC_ForceActive = set_TC_ForceActive;
	TC_ForceInactive = set_TC_ForceInactive;

}


// Called to setup the GPIO_Handler
void GPIO_Handler_init(gpio_instance* pin, GPIO_mode GPIO_mode_val, bool pin_default_state_val)
{
	pin->pin_mode = GPIO_mode_val;
	pin->pin_default_state = pin_default_state_val;
	pin->pin_pulse_counter = 0;
	pin->pin_period_counter = 0;
	pin->pin_period = 0;
	pin->pin_duration = 0;

	if(GPIO_mode_val == DISABLED) return;
	if(isHighSpeed(pin))
	{

		return;
	}


	if(isInput(pin))
	{
		//if input record startup value so we know when it has toggled
		settings_writeInt(pin->pin_state_setting, pin->get_Input_GPIO());
	}else{
		// if output record default state and set the default output of that pin
		pin->set_Output_GPIO(pin_default_state_val);
	}
}

// Should be called each tick (typically 1ms)
void GPIO_Handler_tick(gpio_instance* pin)
{
	if(pin->pin_mode == DISABLED) return;

	// handle input pins separately
	if(isInput(pin))
	{
		if(pin->get_Input_GPIO() != pin->pin_state_setting->value){
			//pin has toggled and we need to toggle the pump state
			settings_writeInt(settings_getHandle()->pumpEnable, !(settings_getHandle()->pumpEnable->value));
			//record new state
			settings_writeInt(pin->pin_state_setting, !(pin->pin_state_setting->value));
		}

		return;
	}


	// handle output pins

	// high speed pins use timer counters and need to be handled separately
	if(isHighSpeed(pin))
	{
		// only do things if there has been an input for pulses
		if(!pin->pin_state_setting->updated) return;

		//acknowledge input
		settings_updateNotedInt(pin->pin_state_setting);

		// handle static output separately
		if(pin->pin_state_setting->value == PIN_STATE_FOR_CONST_ON)
		{
			if(pin->pin_default_state) pin->setTimControlMode(TC_ForceInactive);
			else pin->setTimControlMode(TC_ForceActive);
			return;
		}
		if(pin->pin_state_setting->value == PIN_STATE_FOR_CONST_OFF)
		{
			if(pin->pin_default_state) pin->setTimControlMode(TC_ForceActive);
			else pin->setTimControlMode(TC_ForceInactive);
			return;
		}

		// handle pulse generation separately
		//nothing to do if there is no pulse length
		if (pin->pin_pulse_duration_setting->value <=0 )
		{
			settings_writeInt(pin->pin_state_setting, 0);	//acknowledge we have received the command
			settings_updateNotedInt(pin->pin_state_setting); //prevent retriggering again due to changing the value
			return;
		}

		// set positive or negative pulse generation depening on the default state
		if(pin->pin_default_state) pin->setTimControlMode(TC_PWMMode1);
		else pin->setTimControlMode(TC_PWMMode2);

		// use local variables
		pin->pin_period = pin->pin_pulse_period_setting->value;
		pin->pin_duration = pin->pin_period - pin->pin_pulse_duration_setting->value;
		if (pin->pin_duration <= 0) pin->pin_duration = 1;	// period has to be >= pulse length so extend the period if needed

		//start pulses
		pin->startTimPulse((uint8_t)pin->pin_state_setting->value - 1,
									pin->pin_period - 1,
									pin->pin_duration);
		settings_writeInt(pin->pin_state_setting, 0);	//acknowledge we have received the command
		settings_updateNotedInt(pin->pin_state_setting); //prevent retriggering again due to changing the value
		return;
	}

	// normal GPIO pulse generation
	// static states and setup for pulses

	// only do things if there has been an input for pulses
	if(pin->pin_state_setting->updated)
	{

		settings_updateNotedInt(pin->pin_state_setting);

		//handle the static states first
		if(pin->pin_state_setting->value == PIN_STATE_FOR_CONST_ON)
		{
			pin->set_Output_GPIO(!pin->pin_default_state);
			pin->pin_period_counter = 0;
			pin->pin_pulse_counter = 0;
			return;
		}
		if(pin->pin_state_setting->value == PIN_STATE_FOR_CONST_OFF)
		{
			pin->set_Output_GPIO(pin->pin_default_state);
			pin->pin_period_counter = 0;
			pin->pin_pulse_counter = 0;
			return;
		}

		if (pin->pin_pulse_duration_setting->value <=0 ) return;	//nothing to do if there is no pulse length

		// reset and use local variables
		pin->pin_pulse_counter = pin->pin_state_setting->value;
		pin->pin_period_counter = 0;
		pin->pin_period = pin->pin_pulse_period_setting->value;
		pin->pin_duration = pin->pin_pulse_period_setting->value - pin->pin_pulse_duration_setting->value;
		if (pin->pin_duration <= 0) pin->pin_duration = 1;	// capping the fully on state

		settings_writeInt(pin->pin_state_setting, 0);	//acknowledge we have received the command
		settings_updateNotedInt(pin->pin_state_setting); //prevent retriggering again due to changing the value
	}

	// actually generating the gpio pulses

	//return if there are no pulses to make
	if (pin->pin_period_counter == 0 && pin->pin_pulse_counter == 0) return;

	// we start a new cycle, reset counter and set pin ON
	// if the state is 0 then we have finished and therefore no need to start another cycle
	if (pin->pin_period_counter == 0)
	{
		pin->pin_pulse_counter -= 1;
		pin->pin_period_counter = pin->pin_period;
		pin->set_Output_GPIO(!pin->pin_default_state);
	}

	// when we reach the end of the period the pulse is over and we go back to OFF state
	if (pin->pin_period_counter == pin->pin_duration)
	{
		pin->set_Output_GPIO(pin->pin_default_state);
	}

	pin->pin_period_counter--;
}


//helper function
bool isInput(gpio_instance* pin)
{
	return pin->pin_mode == INPUT_NO_PULL ||
			pin->pin_mode == INPUT_PULL_UP ||
			pin->pin_mode == INPUT_PULL_DOWN;
}

//helper function
bool isHighSpeed(gpio_instance* pin)
{
	return pin->pin_mode == OUTPUT_HIGH_SPEED_DEF_OFF ||
			pin->pin_mode == OUTPUT_HIGH_SPEED_DEF_ON;
}
