/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "i2cMasterHandler.h"

typedef struct device
{
    i2c_device dev;
    i2c_request pendingRequest;
    uint16_t minTicksTillService;
} device;

static i2c_setup set;
static device devices[I2C_MAX_DEVICES];
static uint8_t activeDevices;
static uint8_t currentDevice;

static void serviceNextServicableDevice();
static void serviceDevice(uint8_t deviceIdx);
static void handlePendingRequest(uint8_t deviceIdx, bool success);
static void decrementDeviceTicksTillService();
static void incrementCurrentDevice();

void i2cMaster_init(i2c_setup setup, bool * wasDeviceFound)
{
    set = setup;
    activeDevices = 0;
    currentDevice = 0;
    
    // Find out which devices are on the bus
    // If they're on the bus, add them to our list of devices
    for(uint8_t deviceIdx=0; deviceIdx<setup.deviceCount; deviceIdx++)
    {
    	wasDeviceFound[deviceIdx] = setup.checkDeviceIsAvailable(setup.devices[deviceIdx].i2cAddr);
        if(wasDeviceFound[deviceIdx])
        {
            devices[activeDevices].dev = setup.devices[deviceIdx];
            devices[activeDevices].minTicksTillService = 0;
            activeDevices++;
        }
    }
}

void i2cMaster_forceAddDevice(i2c_device device, uint16_t ticksTillFirstService)
{
	devices[activeDevices].dev = device;
	devices[activeDevices].minTicksTillService = ticksTillFirstService;
	activeDevices++;
}

void i2cMaster_tick(i2c_state state)
{
    if(activeDevices == 0)
        return;  // Don't bother if no devices on bus

    decrementDeviceTicksTillService();

    switch(state)
    {
        case I2C_BUSY:
            break;

        case I2C_IDLE:
            serviceNextServicableDevice();
            break;

        case I2C_SUCCESS:
            handlePendingRequest(currentDevice, true);
            incrementCurrentDevice();
            serviceNextServicableDevice();
            break;

        case I2C_FAILURE:
            handlePendingRequest(currentDevice, false);
            incrementCurrentDevice();
            serviceNextServicableDevice();
            break;
    }
}

static void serviceNextServicableDevice()
{
    // Check activeDevices number of devices, to see if
    // any can be serviced
    for(uint8_t _=0; _<activeDevices; _++)
    {
        if(devices[currentDevice].minTicksTillService == 0)
        {
            serviceDevice(currentDevice);
            break;  // Break out of the for loop
        }

        incrementCurrentDevice();
    }
}

static void serviceDevice(uint8_t deviceIdx)
{
    i2c_request request = devices[deviceIdx].dev.serviceI2CDevice();

    if(request.type == I2C_READ)
    {
        set.readBytes(devices[currentDevice].dev.i2cAddr, 
                      request.buf, 
                      request.count);
    }
    else
    {
        set.sendBytes(devices[currentDevice].dev.i2cAddr, 
                      request.buf, 
                      request.count);
    }

    devices[deviceIdx].pendingRequest = request;
}

static void handlePendingRequest(uint8_t deviceIdx, bool success)
{
    devices[deviceIdx].minTicksTillService = 
        devices[deviceIdx].pendingRequest.requestCompleteCallback(success);
}

static void decrementDeviceTicksTillService()
{
    for(uint8_t deviceIdx=0; deviceIdx<activeDevices; deviceIdx++)
    {
        if(devices[deviceIdx].minTicksTillService > 0)
            devices[deviceIdx].minTicksTillService--;
    }
}

static void incrementCurrentDevice()
{
    currentDevice = (currentDevice + 1) % activeDevices;
}

