/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

//****************************************************************************/
// i2cMasterHandler is a common code module used to arbitrate access to an
// I2C bus where 'we' are the master. 
//
// Devices are registered in the i2c_setup struct, which is then passed through
// i2cMaster_init().
//
// Upon init i2cMasterHandler checks which devices exist on the bus.
// Devices which exist on the bus are serviced in a round-robin fashion. 
// A 'service' constitutes allowing one read or write on the I2C bus. After
// which success or failure is returned in a callback (requestCompleteCallback).
//
// requestCompleteCallback should return the minimum number of ticks till the
// device should be serviced again. 
//****************************************************************************/

#ifndef I2CMASTERHANDLER_H
#define I2CMASTERHANDLER_H

#define I2C_MAX_DEVICES 6

#include <stdbool.h>
#include <stdint.h>

//
// Structs for representing interactions with devices on the bus
//

typedef enum { I2C_READ, I2C_WRITE } i2c_requestType;

typedef struct i2c_request
{
    i2c_requestType type;  // The type of the request (read or write)
    uint8_t count;  // Number of bytes to transfer
    uint8_t* buf;  // Pointer to buffer to read from, or write too

    // Function to call when request complete
    // Should return the minimum number of milliseconds before it can be
    // serviced again
    uint16_t (*requestCompleteCallback)(bool success);  
} i2c_request;

typedef struct i2c_device
{
    uint8_t i2cAddr;
    i2c_request (*serviceI2CDevice)();
} i2c_device;

//
// Struct passed on init, which includes function pointers to hook into the 
// hardware, and a pointer to an array of devices potentially on the bus, which
// we should attempt to service
//

typedef struct i2c_setup
{
    // Function points to hook into hardware
    bool (*checkDeviceIsAvailable)(uint8_t addr);
    void (*sendBytes)(uint8_t addr, uint8_t* buf, uint8_t bytesToSend);
    void (*readBytes)(uint8_t addr, uint8_t* buf, uint8_t bytesToRead);

    i2c_device* devices;
    uint8_t deviceCount;
} i2c_setup;

typedef enum 
{ 
    I2C_IDLE,  // The I2C bus is idle
    I2C_BUSY,  // The I2C bus is busy performing a transaction
    I2C_SUCCESS,  // The I2C bus idle; the requested transaction succeeded
    I2C_FAILURE  // The I2C bus idle; the requested transaction failed 
} i2c_state;

// Called to setup the i2cMasterHandler
// returns a bool array of the devices that were found
void i2cMaster_init(i2c_setup setup, bool * wasDeviceFound);

// Some devices dont respond until a certain condition is met or require a minimum first delay
void i2cMaster_forceAddDevice(i2c_device device, uint16_t ticksTillFirstService);

// Should be called each tick (typically 1ms)
void i2cMaster_tick(i2c_state state);

#endif // I2CMASTERHANDLER_H

