/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "i2cMs563702BA02Handler.h"

typedef enum
{
    RESET,
    SELECTING_COEFF,
    READING_COEFF,
    CONVERTING_PRESSURE,
    SELECTING_PRESSURE,
    READING_PRESSURE,
    CONVERTING_TEMPERATURE,
    SELECTING_TEMPERATURE,
    READING_TEMPERATURE
} deviceState;

static i2c_device device;

static deviceState state;
static uint8_t selectedCoeff;

static uint8_t buf[4];
static uint32_t pressureData;
static uint32_t temperatureData;
static uint32_t pressureUbar;
static uint16_t coeffs[8];  // Extra slot to support checkCrc()

static i2c_request service();
static uint16_t callback(bool success);
static uint32_t calculatePressureUbar();
static bool checkCrc();

i2c_device ms563702ba03_getDevice()
{
    pressureData = 0;
    temperatureData = 0;
    selectedCoeff = 0;
    state = RESET;

    pressureUbar = 1013000;

    device.i2cAddr = MS563702BA03_I2C_ADDR;
    device.serviceI2CDevice = service;

    return device;
}

uint32_t ms563702ba03_getPressureuBar()
{
    return pressureUbar;
}

static i2c_request service()
{
    i2c_request request;
    request.buf = buf;
    request.requestCompleteCallback = callback;

    switch(state)
    {
        case RESET:
            buf[0] = 0x1E;
            request.type = I2C_WRITE;
            request.count = 1;
            break;

        case SELECTING_COEFF:
            buf[0] = 0xA0 | (selectedCoeff << 1);

            request.type = I2C_WRITE;
            request.count = 1;
            break;

        case READING_COEFF:
            request.type = I2C_READ;
            request.count = 2;
            break;

        case CONVERTING_PRESSURE:
            buf[0] = 0x4A;
            request.type = I2C_WRITE;
            request.count = 1;
            break;

        case SELECTING_PRESSURE:
            buf[0] = 0;
            request.type = I2C_WRITE;
            request.count = 1;
            break;

        case READING_PRESSURE:
            request.type = I2C_READ;
            request.count = 3;
            break;

        case CONVERTING_TEMPERATURE:
            buf[0] = 0x5A;
            request.type = I2C_WRITE;
            request.count = 1;
            break;

        case SELECTING_TEMPERATURE:
            buf[0] = 0;
            request.type = I2C_WRITE;
            request.count = 1;
            break;

        case READING_TEMPERATURE:
            request.type = I2C_READ;
            request.count = 3;
            break;
    }

    return request;
}

static uint16_t callback(bool success)
{
    switch(state)
    {
        case RESET:
            if(!success)
                return 0;

            state = SELECTING_COEFF;
            return 10;

        case SELECTING_COEFF:
            if(success)
                state = READING_COEFF;
            return 0;

        case READING_COEFF:
            if(!success)  
                return 0;

            coeffs[selectedCoeff] = buf[1] + ((uint16_t)buf[0] << 8);

            if(selectedCoeff < 6)  // If we've not read everything
            {
                selectedCoeff++;
                state = SELECTING_COEFF;
            }
            else if(checkCrc())
            {
                state = CONVERTING_PRESSURE;
            }
            else  // Retry reading all of the coeffs if bad CRC
            {
                selectedCoeff = 0;
                state = RESET;
            }

            return 0;

        case CONVERTING_PRESSURE:
            if(!success)
                return 0;

            state = SELECTING_PRESSURE;
            return 25;

        case SELECTING_PRESSURE:
            if(success)
                state = READING_PRESSURE;

            return 1; // need some delay as with streaming mode messing up the 1khz cycle time we could send the command too quickly

        case READING_PRESSURE:
            if(success)
            {
                pressureData = (((uint32_t)buf[0]) << 16) + 
                               (((uint32_t)buf[1]) << 8) +
                               buf[2];
            }
            // we always want to move to the next conversion.
            // If we get unsuccessful read and we repeat the read the sensor NACKs as it's not expecting a read (and this repeats forever)
            state = CONVERTING_TEMPERATURE;
            return 1; // need some delay as with streaming mode messing up the 1khz cycle time we could send the command too quickly

        case CONVERTING_TEMPERATURE:
            if(!success)
                return 0;

            state = SELECTING_TEMPERATURE;
            return 25;

        case SELECTING_TEMPERATURE:
            if(success)
                state = READING_TEMPERATURE;

            return 1; // need some delay as with streaming mode messing up the 1khz cycle time we could send the command too quickly

        case READING_TEMPERATURE:
            if(success)
            {
                temperatureData = ((uint32_t)buf[0] << 16) + 
                                  ((uint32_t)buf[1] << 8) +
                                  buf[2];

                pressureUbar = calculatePressureUbar();
            }
            // we always want to move to the next conversion.
            // If we get unsuccessful read and we repeat the read the sensor NACKs as it's not expecting a read (and this repeats forever)
            state = CONVERTING_PRESSURE;
            return 1; // need some delay as with streaming mode messing up the 1khz cycle time we could send the command too quickly

        default:
        	return 0;
    }
}

static uint32_t calculatePressureUbar()
{
    // Calculate compensated pressure value
    // Uses maths from MS563702BA03 datasheet
    int32_t dT = temperatureData - (((uint32_t)coeffs[5]) << 8);

    int32_t TEMP = 2000 + (((int64_t)dT)*coeffs[6]) >> 23;

    int64_t OFF = (((int64_t)coeffs[2]) << 17) + 
        ((((int64_t)coeffs[4]) * dT) >> 6);

    int64_t SENS = (((int64_t)coeffs[1]) << 16) + 
        ((((int64_t)coeffs[3]) * dT) >> 7);

    int32_t P = (((pressureData * SENS) >> 21) - OFF) >> 15;

    return P * 10;
}

//
// Check CRC is modified from the example given in the datasheet for the sensor
//
static bool checkCrc()
{
    // Store the CRC read from the sensor
    uint8_t crc = (coeffs[0] >> 12) & 0b1111;

    uint8_t byteIdx; // simple byteIdxer
    uint16_t remained=0; // crc reminder

    coeffs[0]=((coeffs[0]) & 0x0FFF); // CRC byte is replaced by 0
    coeffs[7]=0; // Subsidiary value, set to 0

    for (byteIdx=0; byteIdx<16; byteIdx++) // operation is performed on bytes
    { 
        if (byteIdx%2 == 1) 
            remained ^= (uint16_t)coeffs[byteIdx>>1] & 0x00FF;
        else 
            remained ^= (uint16_t)coeffs[byteIdx>>1] >> 8;

        for (uint8_t n_bit=8; n_bit>0; n_bit--)
        {
            if(remained & (0x8000)) 
                remained = (remained << 1) ^ 0x3000;
            else 
                remained = (remained << 1);
        }
    }

    remained= ((remained >> 12) & 0x000F); // final 4-bit reminder is CRC code

    // Compare the calculated CRC with that from the sensor
    return remained == crc;
}

