/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "i2cSensirionFlowSensor.h"

typedef enum deviceState
{
    SEND_READ_CMD_A,
    SEND_READ_CMD_B,
    READ_ID,
    SEND_START_CMD,
    READING_FLOW
} deviceState;

static i2c_request service();
static uint16_t callback(bool success);
static bool extractFlowSensorFromBuf();
static bool extractFlowFromBuf();
static bool checkCrc(uint8_t msbyte, uint8_t lsbyte, uint8_t receivedCrc);

static uint8_t buf[6];
static deviceState state;
static float scalingFactor;
static int16_t offsetFactor;
static float flow;

i2c_device flowSensor_getDevice()
{

    i2c_device device;
    state = SEND_READ_CMD_A;
    scalingFactor = 1000;
    offsetFactor = 0;
    flow = 0;

    device.i2cAddr = SENSIRION_FLOW_SENSOR_ADDR_SLF3S_LD20;
    device.serviceI2CDevice = service;

    return device;
}

i2c_device flowSensor_air_getDevice()
{

    i2c_device device;
    state = SEND_READ_CMD_A;
    scalingFactor = 1000;
    offsetFactor = 0;
    flow = 0;

    device.i2cAddr = SENSIRION_FLOW_SENSOR_ADDR_SFM6000D_SFC6000D;
    device.serviceI2CDevice = service;

    return device;
}

float flowSensor_getFlow_mLperMin()
{
    return flow;
}

float flowSensor_getFlow_uLperMin()
{
    return (flow * 1000.0);
}

float flowSensor_getFlow_nLperMin()
{
	return (flow * 1000.0 * 1000.0);
}

static i2c_request service()
{
    i2c_request request;

    switch(state)
    {
        case SEND_READ_CMD_A:
            buf[0] = 0x36;
            buf[1] = 0x7C;
            request.type = I2C_WRITE;
            request.count = 2;
            break;

        case SEND_READ_CMD_B:
            buf[0] = 0xE1;
            buf[1] = 0x02;
            request.type = I2C_WRITE;
            request.count = 2;
            break;

        case READ_ID:
            request.type = I2C_READ;
            request.count = 6;
            break;

        case SEND_START_CMD:
            buf[0] = 0x36;
            buf[1] = 0x08;
            request.type = I2C_WRITE;
            request.count = 2;
            break;

        case READING_FLOW:
            request.type = I2C_READ;
            request.count = 3;
            break;
    }

    request.buf = buf;
    request.requestCompleteCallback = callback;

    return request;
}

static uint16_t callback(bool success)
{
    if(!success)
        return 10;  // If we fail, hold off on retrying for 10 ticks

    switch(state)
    {
        case SEND_READ_CMD_A:
            state = SEND_READ_CMD_B;
            return 0;

        case SEND_READ_CMD_B:
            state = READ_ID;
            return 0;

        case READ_ID:
            if(extractFlowSensorFromBuf())
            {
                state = SEND_START_CMD;
                return 0;
            }
            // If CRC is bad, wait 10 ticks before retrying
            return 10;

        case SEND_START_CMD:
            state = READING_FLOW;
            return 0;

        case READING_FLOW:
            extractFlowFromBuf();
            return 5;
    }

    return 0;  // We shouldn't ever reach here
}

static bool extractFlowSensorFromBuf()
{
    if(!checkCrc(buf[0], buf[1], buf[2]) ||
       !checkCrc(buf[3], buf[4], buf[5]))
    {
        return false;
    }

    // Don't include last byte of product ID, as this is the product revision
    // and can change part to part
    uint32_t id = (buf[0] << 24) +
                  (buf[1] << 16) +
                  (buf[3] << 8);

    //last two hex digits of the ID are the version number which we want to ignore
    switch(id)
    {
        case 0x07030200: 
            //FLOW_SENSOR_SLF3S_1300F;
        	offsetFactor = 0;
            scalingFactor = 500;
            break;

        case 0x07030300: 
            //FLOW_SENSOR_SLF3S_0600F;
        	offsetFactor = 0;
            scalingFactor = 10000;
            break;

        case 0x07010200: 
            //FLOW_SENSOR_LD20_2600B;
        	offsetFactor = 0;
            scalingFactor = 1200;
            break;

        case 0x07010300:
            //FLOW_SENSOR_LD20_0600L;
        	offsetFactor = 0;
            scalingFactor = 72000;
            break;




        case 0x06020100:
			//SFC6000D-50slm;
        	offsetFactor = -28672;
			scalingFactor = 1.024;
			break;

        case 0x06020200:
			//SFC6000D-20slm;
        	offsetFactor = -28672;
			scalingFactor = 2.560;
			break;

        case 0x06020400:
			//SFC6000D-5slm;
        	offsetFactor = -28672;
			scalingFactor = 10.240;
			break;

        case 0x06021100:
			//SFM6000D-50slm;
        	offsetFactor = -28672;
			scalingFactor = 1.024;
			break;

        case 0x06021200:
			//SFM6000D-20slm;
        	offsetFactor = -28672;
			scalingFactor = 2.560;
			break;

        case 0x06021400:
			//SFM6000D-5slm;
        	offsetFactor = -28672;
			scalingFactor = 10.240;
			break;

        default:
            // If we pass the CRC check, but don't recognise the sensor,
            // return true as we don't need to try to re-read the ID, 
            // but keep the sensor as unknown
            // Seems better to try to read from the unknown sensor than to just
            // give up...
        	offsetFactor = 0;
            scalingFactor = 1000;  // Some default value...
            break;
    }

    return true;
}

static bool extractFlowFromBuf()
{
    if(!checkCrc(buf[0], buf[1], buf[2])) 
        return false;

    // we are getting a signed integer so just adding the two bufs together doesnt work
    int16_t temp = 0;
    temp |= buf[0] << 8;
    temp |= buf[1];

    //subtract the offset here as that might cause int16 overflow
    float val = (float) (temp-offsetFactor);
    val /= scalingFactor;

    flow = val;

    return true;
}

// Based on example code from
// hLEE://www.sunshine2k.de/articles/coding/crc/understanding_crc.html#ch21
static bool checkCrc(uint8_t msbyte, uint8_t lsbyte, uint8_t receivedCrc)
{
    // Generator and initial CRC value taken from Sensirion datasheet
    const uint8_t generator = 0x31;
    uint8_t crc = 0xFF; 
    uint8_t bytes[2];

    bytes[0] = msbyte;
    bytes[1] = lsbyte;

    for(uint8_t idx=0; idx<2; idx++)
    {
        uint8_t currByte = bytes[idx];

        crc ^= currByte; /* XOR-in the next input byte */

        for (int i = 0; i < 8; i++)
        {
            if ((crc & 0x80) != 0)
            {
                crc = (uint8_t)((crc << 1) ^ generator);
            }
            else
            {
                crc <<= 1;
            }
        }
    }

    return crc == receivedCrc;
}

