/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "i2cSensirionFlowSensor2.h"

typedef enum deviceState
{
	SEND_READ_RESOLUTION,	//read and set the resolution of the sensor
	READ_RESOLUTION,
	SET_RESOLUTION,
	SEND_READ_SCALE_FACTOR_AND_MEAS_UNIT,	//get scale and measurement unit
    READ_SCALE_FACTOR_AND_MEAS_UNIT,
	TRIGGER_MEASUREMENT,
	READ_RESULTS,
} deviceState;

static i2c_request service();
static uint16_t callback(bool success);
static bool extractFlowFromBuf();
static bool getScaleFactorAndMeasUnitFromBuf();
static bool getAdvancedUserRegisterFromBuf();
static bool checkCrc(uint8_t msbyte, uint8_t lsbyte, uint8_t receivedCrc);

static uint8_t buf[6];
static uint16_t advancedUserRegister;
static deviceState state;
static float scalingFactor;
static float measUnitScalingFactor;
static float flow;	// in ml/min

#define RESOLUTION_MASK 0xF1FF
#define RESOLUTION_VALUE 0b010 //11bits ~2.5ms
#define RESOLUTION_POSITION 9

#define MASTER_HOLD_MASK 0b10

i2c_device flowSensor2_getDevice()
{
    i2c_device device;
    state = SEND_READ_RESOLUTION;
    scalingFactor = 1000;
    measUnitScalingFactor = 1;
    flow = 0;

    device.i2cAddr = SENSIRION_FLOW_SENSOR2_ADDR;
    device.serviceI2CDevice = service;

    return device;
}

float flowSensor2_getFlow_mLperMin()
{
    return flow;
}

float flowSensor2_getFlow_uLperMin()
{
	return (flow * 1000.0);
}

float flowSensor2_getFlow_nLperMin()
{
	return (flow * 1000.0 * 1000.0);
}

static i2c_request service()
{
    i2c_request request;

    switch(state)
    {
    	case SEND_READ_RESOLUTION:
    		buf[0] = 0xE5;
    		request.type = I2C_WRITE;
			request.count = 1;
			break;

    	case READ_RESOLUTION:
			request.type = I2C_READ;
			request.count = 3;
			break;

    	case SET_RESOLUTION:
    		buf[0] = 0xE4;
			buf[1] = advancedUserRegister >> 8;
			buf[2] = advancedUserRegister & 0xFF;
			request.type = I2C_WRITE;
			request.count = 3;
    		break;

        case SEND_READ_SCALE_FACTOR_AND_MEAS_UNIT:
            buf[0] = 0xFA;
            buf[1] = 0x2B;
            buf[2] = 0x60;
            request.type = I2C_WRITE;
            request.count = 3;
            break;

        case READ_SCALE_FACTOR_AND_MEAS_UNIT:
            request.type = I2C_READ;
            request.count = 6;
            break;

		case TRIGGER_MEASUREMENT:
			buf[0] = 0xF1;
			request.type = I2C_WRITE;
			request.count = 1;
			break;

		case READ_RESULTS:
			request.type = I2C_READ;
			request.count = 3;
			break;
    }

    request.buf = buf;
    request.requestCompleteCallback = callback;

    return request;
}

static uint16_t callback(bool success)
{
    if(!success)
        return 10;  // If we fail, hold off on retrying for 10 ticks

    switch(state)
    {
        case SEND_READ_RESOLUTION:
            state = READ_RESOLUTION;
            return 0;

        case READ_RESOLUTION:
        	if(getAdvancedUserRegisterFromBuf())
			{
        		state = SET_RESOLUTION;
        		return 0;
			}
        	// If CRC is bad, wait 10 ticks before retrying
			return 10;

        case SET_RESOLUTION:
        	state = SEND_READ_SCALE_FACTOR_AND_MEAS_UNIT;
			return 0;

        case SEND_READ_SCALE_FACTOR_AND_MEAS_UNIT:
			state = READ_SCALE_FACTOR_AND_MEAS_UNIT;
			return 0;

        case READ_SCALE_FACTOR_AND_MEAS_UNIT:
        	if(getScaleFactorAndMeasUnitFromBuf())
        	{
        		state = TRIGGER_MEASUREMENT;
        		return 0;
        	}
        	// If CRC is bad, wait 10 ticks before retrying
			return 10;

        case TRIGGER_MEASUREMENT:
        	state = READ_RESULTS;
        	return 0;

        case READ_RESULTS:
			if(buf[0]==0xFF && buf[1]==0xFF && buf[2]==0xFF){
				// The first time we read this will actually trigger the sensor measurement and return 0xFFFFFF
				// This also happens if the measurement is not ready yet
				return 3;
			}
			else{
				// if we get data process it and trigger another read
				extractFlowFromBuf();	// if CRC is bad we wont process data
				state = TRIGGER_MEASUREMENT;
				return 0;
			}

    }

    return 0;  // We shouldn't ever reach here
}

static bool getAdvancedUserRegisterFromBuf()
{
	if(!checkCrc(buf[0], buf[1], buf[2]))
			return false;

	advancedUserRegister |= buf[0] << 8;
	advancedUserRegister |= buf[1];

	advancedUserRegister &= RESOLUTION_MASK; //clear resolution bits
	advancedUserRegister |= RESOLUTION_VALUE << RESOLUTION_POSITION; //write the required resolution

	advancedUserRegister &= ~(MASTER_HOLD_MASK); //disable the master hold by clearing the bit

	return true;
}

static bool getScaleFactorAndMeasUnitFromBuf()
{
	if(!checkCrc(buf[0], buf[1], buf[2]))
		return false;

	if(!checkCrc(buf[3], buf[4], buf[5]))
			return false;

	// we are getting a signed integer so just adding the two bufs together doesnt work
	uint16_t temp = 0;
	temp |= buf[0] << 8;
	temp |= buf[1];
	scalingFactor = (float) temp;

	// we are getting a signed integer so just adding the two bufs together doesnt work
	temp = 0;
	temp |= buf[3] << 8;
	temp |= buf[4];

	switch(temp)
	{
	case 0x0843:
		// nl/min
		measUnitScalingFactor = 1000000.0;
		break;

	case 0x0844:
		// ul/min
		measUnitScalingFactor = 1000.0;
		break;

	case 0x0845:
		// ml/min
		measUnitScalingFactor = 1.0;
		break;

	case 0x0834:
		// ul/sec
		measUnitScalingFactor = 16.6667;
		break;

	case 0x0855:
		// ml/h
		measUnitScalingFactor = 60.0;
		break;

	default:
		// If we pass the CRC check, but don't recognise the sensor,
		// return true as we don't need to try to re-read the ID,
		// but keep the sensor as unknown
		// Seems better to try to read from the unknown sensor than to just
		// give up...
		measUnitScalingFactor = 1;	//some value
		break;
	}

	return true;
}


static bool extractFlowFromBuf()
{
    if(!checkCrc(buf[0], buf[1], buf[2])) 
        return false;

    // we are getting a signed integer so just adding the two bufs together doesnt work
    int16_t temp = 0;
    temp |= buf[0] << 8;
    temp |= buf[1];
    float val = (float) temp;
    val = val / (scalingFactor * measUnitScalingFactor);

    flow = val;

    return true;
}

// Based on example code from (for two bytes of data)
// https://sensirion.com/media/documents/CBEB368E/61FAB1B9/LQ_AN_LiquidFlowSensors_ImplementationGuideToI2C_EN_D1.pdf
#define POLYNOMIAL 0x131
static bool checkCrc(uint8_t msbyte, uint8_t lsbyte, uint8_t receivedCrc)
{
	//for two bytes of data
	uint8_t calc_crc = 0;
	uint8_t data[2];

	data[0] = msbyte;
	data[1] = lsbyte;

	for (uint8_t byteCtr = 0; byteCtr < 2; ++byteCtr){
		calc_crc ^= (data[byteCtr]);
		for (int8_t i = 8; i > 0; --i){
			if (calc_crc & 0x80) { calc_crc = (calc_crc << 1) ^ POLYNOMIAL;}
			else { calc_crc = (calc_crc << 1); }
		}
	}

	return calc_crc == receivedCrc;
}

