/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "i2cSlaveHandler.h"
#include "Settings.h"

typedef enum 
{
    IDLE,
    RECEIVING_HEADER,
    RECEIVING_INT,
    RECEIVING_FLOAT,
    WAIT_TO_SEND_INT,
    WAIT_TO_SEND_FLOAT
} i2cState;

static const uint8_t readMask = 1 << 7;

static i2cSlaveHandler_settings setup;

static volatile i2cState state;
static uint8_t headerByte;
static setting_int* intSettingPtr;
static setting_float* floatSettingPtr;
static int16_t intData;
static float floatData;

static uint8_t streamBuf[32];
static uint8_t streamBufLen;
static bool streamingEnabled = false;


void i2cSlaveHandler_init(i2cSlaveHandler_settings settings)
{
    setup = settings;
    state = IDLE;
    streamingEnabled = false;
}

void i2cSlaveHandler_masterStartedMessage(bool isWrite)
{
    if(isWrite)
    {
        setup.receiveBytesFromMaster(&headerByte, 1);
        state = RECEIVING_HEADER;
    }
    else
    {
        if(state == WAIT_TO_SEND_INT)
        {
            intData = intSettingPtr->value;
            setup.writeBytesToMaster((uint8_t*)&intData, 2);
            state = IDLE;
        }
        else if(state == WAIT_TO_SEND_FLOAT)
        {
            floatData = floatSettingPtr->value;
            setup.writeBytesToMaster((uint8_t*)&floatData, 4);
            state = IDLE;
        }
        else
        {
        	// Master requests to read without giving a register id
        	// this is either streaming mode
        	// or unexpected read, just return 1 byte of 0

        	if(streamingEnabled){
        		streamBufLen = 0;
        		// Concatenate setting values for stream
				for(uint8_t idx=0; idx<setup.settingsToStreamCount; idx++)
				{
					if(settings_getTypeForSetting(setup.settingsToStream[idx]) == SETTINGS_TYPE_INT)
					{
						int16_t temp_value = settings_getSettingInt(setup.settingsToStream[idx])->value;
						streamBuf [streamBufLen++] 	= ((uint8_t*) &temp_value)[0];
						streamBuf [streamBufLen++] 	= ((uint8_t*) &temp_value)[1];
					}else
					{
						float temp_value = settings_getSettingFloat(setup.settingsToStream[idx])->value;
						streamBuf [streamBufLen++] 	= ((uint8_t*) &temp_value)[0];
						streamBuf [streamBufLen++] 	= ((uint8_t*) &temp_value)[1];
						streamBuf [streamBufLen++] 	= ((uint8_t*) &temp_value)[2];
						streamBuf [streamBufLen++] 	= ((uint8_t*) &temp_value)[3];
					}
				}

				// Calculate the stream checksum
				uint8_t checksum = 0;

				for(uint16_t idx=0; idx<streamBufLen; idx++)
					checksum += streamBuf[idx];

				streamBuf [streamBufLen++] = checksum;

            	setup.writeBytesToMaster((uint8_t*)&streamBuf, streamBufLen);

        	}else
        	{
        		intData = 0;
        		setup.writeBytesToMaster((uint8_t*)&intData, 1);
        	}
            state = IDLE;
        }
    }
}

void i2cSlaveHandler_slaveFinishedReceiving()
{
    switch(state)
    {
        case RECEIVING_HEADER:
            if(headerByte & readMask)  // Master wants to read
            {
                uint8_t settingIdx = headerByte & ~readMask;

                if(settings_getTypeForSetting(settingIdx) == SETTINGS_TYPE_INT)
                {
                    intSettingPtr = settings_getSettingInt(settingIdx);
                    state = WAIT_TO_SEND_INT;
                }
                else
                {
                    floatSettingPtr = settings_getSettingFloat(settingIdx);
                    state = WAIT_TO_SEND_FLOAT;
                }
            }
            else if(settings_getTypeForSetting(headerByte) == SETTINGS_TYPE_INT)
            {
                setup.receiveBytesFromMaster((uint8_t*) &intData, 2);
                state = RECEIVING_INT;
            }
            else
            {
                setup.receiveBytesFromMaster((uint8_t*) &floatData, 4);
                state = RECEIVING_FLOAT;
            }
            break;

        case RECEIVING_INT:
            if(settings_getCRCcheckedForSetting(headerByte)) settings_writeCRCInt(settings_getSettingInt(headerByte), intData, headerByte);
            else settings_writeInt(settings_getSettingInt(headerByte), intData);
            break;

        case RECEIVING_FLOAT:
            settings_writeFloat(settings_getSettingFloat(headerByte), floatData);
            break;

        case IDLE:
        case WAIT_TO_SEND_INT:
        case WAIT_TO_SEND_FLOAT:
            break;  // Do nothing

    }
}

// Call to turn streaming off or on
void i2cSlaveHandler_setStreaming(bool enabled)
{
	streamingEnabled = enabled;
}

