/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "phaseDetector.h"
#include "fastATan2.h"
#include <math.h>

#define LOOKUP_SIZE_POWER (7)
#define LOOKUP_PERIOD (1 << LOOKUP_SIZE_POWER)
#define LOOKUP_SIZE (LOOKUP_PERIOD * 5 / 4)
#define LOOKUP_STEP_SHIFT (32 - LOOKUP_SIZE_POWER)
#define LOOKUP_MAGNITUDE ((1 << 15) - 1)

static int16_t sinLookup[LOOKUP_SIZE];
static int16_t* cosLookup;

static uint32_t sampleRate;
static uint8_t filterFactor;
static uint32_t countStepSize;

// Accumulators, for holding rolling magnitude of the signals in
// quadrature components
static int32_t aSinAccum;
static int32_t aCosAccum;
static int32_t bSinAccum;
static int32_t bCosAccum;

static float calculatePhase();
static int16_t calculateAmplitudeRms(int32_t sinAccum, int32_t cosAccum);

void phaseDetector_init(uint32_t samplingFrequency,
                        uint8_t smoothingFactor, 
                        uint32_t initialFrequencyHz)
{
    sampleRate = samplingFrequency;
    filterFactor = smoothingFactor;
    countStepSize = ((uint64_t)initialFrequencyHz << 32) / sampleRate;

    aSinAccum = 0;
    aCosAccum = 0;
    bSinAccum = 0;
    bCosAccum = 0;

    for(int idx=0; idx<LOOKUP_SIZE; idx++)
    {
        sinLookup[idx] = 
            sinf(idx * M_PI * 2.0 / LOOKUP_PERIOD) * LOOKUP_MAGNITUDE;
    }

    // cos lookup table uses same memory as sinLookup, but just shifted by
    // 1/4 of a period. This saves memory compared to having two seperate tables
    // for sine and cos, but is faster to lookup than just one table where you
    // calculate the offset to get cos (rather than sine)
    cosLookup = sinLookup + (LOOKUP_PERIOD/4);
}

void phaseDetector_setSignalFrequency(uint32_t frequencyHz)
{
    countStepSize = ((uint64_t)frequencyHz << 32) / sampleRate;
}

void phaseDetector_newSamples(int16_t signalASample, int16_t signalBSample)
{
    static uint32_t count = 0;

    count += countStepSize;
    uint16_t idx = count >> LOOKUP_STEP_SHIFT;

    int16_t sineMulti = sinLookup[idx];
    int16_t cosMulti = cosLookup[idx];

    // Apply exponential filter
    aSinAccum -= aSinAccum >> filterFactor;
    aSinAccum += (signalASample * sineMulti) >> 16;

    aCosAccum -= aCosAccum >> filterFactor;
    aCosAccum += (signalASample * cosMulti) >> 16;

    bSinAccum -= bSinAccum >> filterFactor;
    bSinAccum += (signalBSample * sineMulti) >> 16;

    bCosAccum -= bCosAccum >> filterFactor;
    bCosAccum += (signalBSample * cosMulti) >> 16;
}

phaseDetector_update phaseDetector_getUpdate()
{
    phaseDetector_update update;

    update.phaseDegrees = calculatePhase();
    update.amplitudeARms = calculateAmplitudeRms(aSinAccum, aCosAccum);
    update.amplitudeBRms = calculateAmplitudeRms(bSinAccum, bCosAccum);

    return update;
}

static float calculatePhase()
{
    float aAngle = fastATan2(aSinAccum, aCosAccum);
    float bAngle = fastATan2(bSinAccum, bCosAccum);
    float phase = aAngle - bAngle;

    if(phase > 180)
        phase -= 360;
    else if(phase < -180)
        phase += 360;

    return phase;
}

static int16_t calculateAmplitudeRms(int32_t sinAccum, int32_t cosAccum)
{
    int32_t x = sinAccum >> (filterFactor-2);
    int32_t y = cosAccum >> (filterFactor-2);

    return sqrtf((x*x) + (y*y)) / sqrtf(2.0);
}

