/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#ifndef SETTINGS_H
#define SETTINGS_H

#include <stdbool.h>
#include <stdint.h>

typedef enum
{
	OUTPUT_HIGH_SPEED_DEF_ON = 0U,
	OUTPUT_HIGH_SPEED_DEF_OFF,
	OUTPUT_LOW_SPEED_DEF_ON,
	OUTPUT_LOW_SPEED_DEF_OFF,
	INPUT_NO_PULL,
	INPUT_PULL_UP,
	INPUT_PULL_DOWN,
	DISABLED,

	GPIO_MODE_COUNT
} GPIO_mode;

typedef enum
{
    SETTINGS_MODE_MANUAL = 0,
    SETTINGS_MODE_PID,
    SETTINGS_MODE_BANGBANG,

    SETTINGS_MODE_COUNT
} settings_mode;

typedef enum
{
    SETTINGS_SOURCE_SETVAL = 0,
    SETTINGS_SOURCE_ANA1,
    SETTINGS_SOURCE_ANA2,
    SETTINGS_SOURCE_ANA3,
    SETTINGS_SOURCE_FLOW,
    SETTINGS_SOURCE_PRESSURE,

    SETTINGS_SOURCE_COUNT
} settings_sources;

typedef enum
{
    SETTINGS_TYPE_INVALID,
    SETTINGS_TYPE_INT,
    SETTINGS_TYPE_FLOAT,
} settings_type;

//int struct
typedef struct
{
    settings_type type;
    int16_t value;
    int16_t minValue;
    int16_t maxValue;
    bool updated;
    bool userWritable;
    bool CRCchecked;
} setting_int;

//float struct
typedef struct
{
    settings_type type;
    float value;
    float minValue;
    float maxValue;
    bool updated;
    bool userWritable;
    bool CRCchecked;
} setting_float;

//struct containing all settings
//packed to 4-byte boundaries, so it can be accessed as a list of pointers
#pragma pack(push, 4)
typedef struct
{
    const setting_int* pumpEnable; //0
    const setting_int* powerLimitMilliwatts; //1

    const setting_int* enableStreamMode; //2

    const setting_float* measDriveVolts; //3
    const setting_float* measDriveMilliamps; //4
    const setting_float* measDriveMilliwatts; //5
    const setting_int* measDriveFreq; //6
    const setting_float* measAna1; //7
    const setting_float* measAna2; //8
    const setting_float* measAna3; //9

    const setting_int* mode; //10

    const setting_int* manualModeSource; //11

    const setting_int* pidModeSetpointSource; //12
    const setting_int* pidModeMeasSource; //13
    const setting_float* pidProportionalCoeff; //14
    const setting_float* pidIntegralCoeff; //15
    const setting_float* pidIntegralLimitCoeff; //16
    const setting_float* pidDifferentialCoeff; //17

    const setting_int* bangBangMeasSource; //18
    const setting_float* bangBangLowerThresh; //19
    const setting_float* bangBangUpperThresh; //20
    const setting_float* bangBangLowerPowerMilliwatts; //21
    const setting_float* bangBangUpperPowerMilliwatts; //22

    const setting_float* setVal; //23
    const setting_float* ana1Offset; //24
    const setting_float* ana1Gain; //25
    const setting_float* ana2Offset; //26
    const setting_float* ana2Gain; //27
    const setting_float* ana3Offset; //28
    const setting_float* ana3Gain; //29

    const setting_int* storeCurrentSettings; //30

    const setting_int* errorCode; //31

    const setting_float* measFlow; //32

    const setting_int* resetPidOnTurnon; //33

    const setting_int* useFrequencyTracking; //34
    const setting_int* manualDriveFrequency; //35

    const setting_int* firmwareVersion; //36
    const setting_int* firmwareType; //37 (device type this firmware is for)
    const setting_int* firmwareMinorVersion; //38

    const setting_float* digitalPressure; //39
    const setting_float* digitalPressureOffset; //40

    const setting_float* measDrivePhase; //41

    const setting_int* driverI2cAddress; 		//42
    const setting_int* communicationInterface; 	//43

	const setting_int* pinModeGPIOA; 			//44
	const setting_int* pinStateGPIOA; 			//45
	const setting_int* pinPulseDurationGPIOA; 	//46
	const setting_int* pinPulsePeriodGPIOA; 	//47

	const setting_int* pinModeGPIOB; 			//48
	const setting_int* pinStateGPIOB; 			//49
	const setting_int* pinPulseDurationGPIOB; 	//50
	const setting_int* pinPulsePeriodGPIOB; 	//51

	const setting_int* pinModeGPIOC; 			//52
	const setting_int* pinStateGPIOC; 			//53
	const setting_int* pinPulseDurationGPIOC; 	//54
	const setting_int* pinPulsePeriodGPIOC; 	//55

	const setting_int* pinStateGPIOD; 			//56

	const setting_int* LED_Colour; 				//57

	const setting_int* DigitalPressureMeasUnit; //58
	const setting_int* FlowMeasUnit; 			//59
} settings;
#pragma pack(pop)


settings* settings_getHandle();

bool settings_writeInt(const setting_int* setting, int newVal);

//special function for writing to CRC protected settings
bool settings_writeCRCInt(const setting_int* setting, int newVal, uint8_t idx);

bool settings_writeFloat(const setting_float* setting, float newVal);

void settings_updateNotedInt(const setting_int* setting);

void settings_updateNotedFloat(const setting_float* setting);

uint16_t settings_getSettingsCount();

settings_type settings_getTypeForSetting(uint16_t settingIdx);

setting_int* settings_getSettingInt(uint16_t settingIdx);

setting_float* settings_getSettingFloat(uint16_t settingIdx);

bool settings_getSettingAsString(uint8_t settingIdx,
                                 uint8_t* settingStr, 
                                 uint8_t* settingStrLen);

bool settings_setSettingAsString(uint8_t settingIdx,
                                 uint8_t* settingStr, 
                                 uint8_t settingStrLen);

bool settings_getCRCcheckedForSetting(uint8_t settingIdx);

uint16_t settings_generateCRCcheck(uint8_t settingIdx, uint16_t value);

#endif // SETTINGS_H
