/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
#include "stringHandler.h"

static uint32_t magnitudes[] =
{
    1,
    10,
    100,
    1000,
    10000,
    100000,
    1000000,
    10000000,
    100000000,
    1000000000,
    2147483648  // int32_t min value
};

static uint8_t getDigitsInInt(uint32_t val);

void stringHandler_intAsString(int32_t val, uint8_t* str, uint8_t* strLen)
{
    bool isNegative = val < 0;
    uint8_t digitsStartIndex = 0;

    // Make value positive, we tack on a negative sign at the end
    if(isNegative)
        val *= -1;

    uint8_t numDigits = getDigitsInInt(val);

    // Add negative sign at the beginning if negative, and shift values to right
    if(isNegative)
    {
        str[0] = '-';
        digitsStartIndex++;
        numDigits++;
    }

    for(uint8_t digitIdx=digitsStartIndex; digitIdx<numDigits; digitIdx++)
    {
        str[digitIdx] = 
            '0' + (val / magnitudes[numDigits - (digitIdx+1)]) % 10;
    }

    *strLen = numDigits;
}

bool stringHandler_stringAsInt(int32_t* val, uint8_t* str, uint8_t strLen)
{
    uint8_t isNegative = 0;
    *val = 0;

    if(str[0] == '-')
        isNegative = 1;

    for(uint8_t digitIdx=isNegative; digitIdx<strLen; digitIdx++)
    {
        if(str[digitIdx] < '0' || str[digitIdx] > '9')
            return false;

        *val *= 10;
        *val += str[digitIdx] - '0';
    }

    if(isNegative)
        *val *= -1;

    return true;
}

void stringHandler_floatAsString(float val, uint8_t* str, uint8_t* strLen)
{
    // If NaN just return "0"
    if(isnan(val))
    {
        str[0] = '0';
        *strLen = 1;
        return;
    }

    // If 5 digits or more, just print as an int
    if(val >= 10000)
    {
        stringHandler_intAsString((int32_t) (val + 0.5), str, strLen);
        return;
    }
    else if(val <= -10000)
    {
        stringHandler_intAsString((int32_t) (val - 0.5), str, strLen);
        return;
    }

    // Add a negative symbol at the start if negative
    bool isNegative = val < 0;

    if(isNegative)
    {
        str[0] = '-';
        str++;
        val *= -1;
    }

    // First part of string formatted as an int
    stringHandler_intAsString((int32_t) val, str, strLen);

    uint8_t decimalPlacesToShow = 5 - *strLen;

    str[(*strLen)++] = '.';

    // Remove integer component of value
    val -= (int32_t) val;

    // Shift appropriate number of decimal places to integer
    val *= magnitudes[decimalPlacesToShow];

    uint8_t zerosAfterDecimalPlace = 
        decimalPlacesToShow - getDigitsInInt((uint32_t) val);

    for(uint8_t count=0; count<zerosAfterDecimalPlace; count++)
        str[(*strLen)++] = '0';

    uint8_t decimalSectionLen;
    stringHandler_intAsString((int32_t)(val + 0.5), 
                              str + *strLen, 
                              &decimalSectionLen);
    *strLen += decimalSectionLen;

    // Work back from the end of the string, removing trailing zeros
    while(str[*strLen-1] == '0' || str[*strLen-1] == '.')
    {
        (*strLen)--;

        if(str[*strLen] == '.')
            break;
    }

    if(isNegative)
        (*strLen)++;
}

bool stringHandler_stringAsFloat(float* val, uint8_t* str, uint8_t strLen)
{
    *val = 0;
    uint8_t idx = 0;
    bool isNegative = str[0] == '-';

    if(isNegative)
    {
        str++;
        strLen--;
    }

    for(; idx<strLen; idx++)
    {
        if(str[idx] < '0' || str[idx] > '9')
        {
            if(str[idx] == '.')
            {
                idx++;
                break;
            }
            else
                return false;
        }

        *val *= 10;
        *val += str[idx] - '0';
    }

    uint8_t fraction = 1;

    for(; idx<strLen; idx++)
    {
        if(str[idx] < '0' || str[idx] > '9')
            return false;

        *val += (float)(str[idx] - '0') / magnitudes[fraction++];
    }

    if(isNegative)
        *val *= -1;

    return true;
}

static uint8_t getDigitsInInt(uint32_t val)
{
    if(val == 0)  // 0 is a special case
        return 1;

    uint8_t numDigits = 0;

    while(val >= magnitudes[numDigits] && numDigits < 10)
        numDigits++;
    
    return numDigits;
}
