/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#include <stdio.h>
#include "uartHandler.h"

enum ParseState
{
    WAITING_FOR_HASH,
    WAITING_FOR_COMMAND_TYPE,
    CAPTURE_READ_IDX,
    CAPTURE_WRITE_IDX,
    CAPTURE_WRITE_VALUE,
};

static enum ParseState state = WAITING_FOR_HASH;

static uartHandler_setup setup;

static uint8_t txBuf[32];
static uint8_t txBufLen;

static uint8_t streamBuf[128];
static uint8_t streamBufLen;

static uint8_t debugBuf[32];
static uint8_t debugBufLen;

static uint16_t streamCountDown;
static bool streamingEnabled = false;
static bool uartSendAttempted = false;

static uint8_t settingBuf[16];
static uint8_t settingBufLen;
static uint16_t settingNum;
static uint8_t settingNumLen = 0;

static void handleStream();
static void handleCommandType(uint8_t byte);
static void handleReadIdx(uint8_t byte);
static void handleWriteIdx(uint8_t byte);
static void handleWriteValue(uint8_t byte);
static void handleUnexpectedByte(uint8_t byte);

void uartHandler_init(uartHandler_setup setupDetails)
{
    setup = setupDetails;
    streamingEnabled = false;
}

void uartHandler_tick()
{
    static uint8_t rxBuf[256];
    static uint8_t rxBufLen;
    uartSendAttempted = false;

    // Try to TX bytes if buffer full (previously failed to send)
    if(streamBufLen > 0)
    {
        if(setup.sendBytes(streamBuf, streamBufLen))
            streamBufLen = 0;
        uartSendAttempted = true;
    }

    // Stream out settings if enabled and countdown reached
    /* Only do if we are not processing an incoming command
     * as both Stream and receiving data use the settingBuf
     * and Stream can overwrite the data from an incoming command
     */
    if (state == WAITING_FOR_HASH)
    {
    	handleStream();
    }

    if(streamBufLen > 0)
        return;

    // Try to TX bytes if buffer full (previously failed to send)
    if(txBufLen > 0 && !uartSendAttempted)
    {
        if(setup.sendBytes(txBuf, txBufLen))
            txBufLen = 0;
        uartSendAttempted = true;
    }

    if(txBufLen > 0)
        return;

    // Get unread bytes, and parse
    if(!setup.getUnreadBytes(rxBuf, &rxBufLen))
    {
        // Send debug string if available
        if(debugBufLen > 0 && !uartSendAttempted)
        {
            if(setup.sendBytes(debugBuf, debugBufLen))
                debugBufLen = 0;
        }

        return;
    }

    for(uint8_t byteIdx=0; byteIdx<rxBufLen; byteIdx++)
    {
        uint8_t byte = rxBuf[byteIdx];

        switch(state)
        {
            case WAITING_FOR_HASH:
                if(byte == '#')
                    state = WAITING_FOR_COMMAND_TYPE;
                break;

            case WAITING_FOR_COMMAND_TYPE:
                handleCommandType(byte);
                break;

            case CAPTURE_READ_IDX:
                handleReadIdx(byte);
                break;

            case CAPTURE_WRITE_IDX:
                handleWriteIdx(byte);
                break;

            case CAPTURE_WRITE_VALUE:
                handleWriteValue(byte);
                break;
        }
    }
}

void uartHandler_setStreaming(bool enabled)
{
    streamingEnabled = enabled;
    streamCountDown = 0;
}

void uartHandler_sendDebug(uint8_t* debugString, 
                           uint8_t debugStringLen)
{
    uint8_t* buf = debugBuf;

    buf[0] = '#';
    buf[1] = 'D';
    buf[2] = '_';

    buf += 3;

    for(uint8_t idx=0; idx<debugStringLen; idx++)
        buf[idx] = debugString[idx];

    buf += debugStringLen;

    buf[0] = '\n';

    debugBufLen = debugStringLen + 4;
}

static void handleStream()
{
    if(!streamingEnabled)
        return;

    if(streamCountDown <= 0)
    {
        streamBuf[0] = '#';
        streamBuf[1] = 'S';
        streamBufLen = 2;

        // Concatenate setting values for stream
        for(uint8_t idx=0; idx<setup.settingsToStreamCount; idx++)
        {
            if(setup.getSetting(setup.settingsToStream[idx],
                                settingBuf, &settingBufLen))
            {
                for(uint8_t c=0; c<settingBufLen; c++)
                {
                    streamBuf[streamBufLen++] = settingBuf[c];
                }
            }
            else
            {
                streamBuf[streamBufLen++] = '0';
            }

            streamBuf[streamBufLen++] = ',';
        }

        // Calculate the stream checksum
        uint8_t checksum = 0;

        for(uint16_t idx=0; idx<streamBufLen; idx++)
            checksum += streamBuf[idx];

        streamBufLen += sprintf(streamBuf + streamBufLen, "%d\n", checksum);

        if(setup.sendBytes(streamBuf, streamBufLen))
            streamBufLen = 0;

        uartSendAttempted = true;

        streamCountDown = setup.ticksPerStream;
    }

    streamCountDown--;
}

static void handleCommandType(uint8_t byte)
{
    if(byte == 'R')
    {
        state = CAPTURE_READ_IDX;
        settingNum = 0;
    }
    else if(byte == 'W')
    {
        state = CAPTURE_WRITE_IDX;
        settingNum = 0;
    }
    else
        handleUnexpectedByte(byte);
}

static void handleReadIdx(uint8_t byte)
{
    if(byte >= '0' && byte <= '9')
    {
        settingNum *= 10;
        settingNum += byte - '0';
        settingNumLen++;
    }
    else if(byte == '\n' && settingNumLen > 0 &&
            settingNum >= 0 && settingNum < 255)
    {
        if(setup.getSetting(settingNum, settingBuf, &settingBufLen))
        {
            settingBuf[settingBufLen] = 0; 

            // Format response with read command
            txBufLen = sprintf(txBuf, "#R%d,%s\n", 
                               settingNum, settingBuf);

            if(setup.sendBytes(txBuf, txBufLen))
                txBufLen = 0;
        }

        settingNumLen = 0;
        state = WAITING_FOR_HASH;
    }
    else
        handleUnexpectedByte(byte);
}

static void handleWriteIdx(uint8_t byte)
{
    if(byte >= '0' && byte <= '9')
    {
        settingNum *= 10;
        settingNum += byte - '0';
        settingNumLen++;
    }
    else if(byte == ',' && settingNumLen > 0 &&
            settingNum >= 0 && settingNum < 255)
    {
        state = CAPTURE_WRITE_VALUE;
        settingNumLen = 0;
        settingBufLen = 0;
    }
    else
        handleUnexpectedByte(byte);
}

static void handleWriteValue(uint8_t byte)
{
    if(byte != '\n')
    {
        settingBuf[settingBufLen++] = byte;
    }
    else if(settingBufLen > 0)
    {
        if(setup.setSetting(settingNum, settingBuf, settingBufLen))
        {
            settingBuf[settingBufLen] = 0;  // Add null terminator for sprintf

            txBufLen = sprintf(txBuf, "#W%d,%s\n", settingNum, settingBuf);

            if(setup.sendBytes(txBuf, txBufLen))
                txBufLen = 0;
        }

        state = WAITING_FOR_HASH;
    }
    else
        handleUnexpectedByte(byte);
}

static void handleUnexpectedByte(uint8_t byte)
{
    if(byte == '#')
        state = WAITING_FOR_COMMAND_TYPE;
    else
        state = WAITING_FOR_HASH;

    settingNumLen = 0;
}

