/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#include "stm32g0xx_hal.h"
#include <stdbool.h>
#include "tim.h"
#include "gpio.h"

#include "pwmDrive.h"

#define PWM_DRIVE_GROUND_TIME_US 3
#define PWM_DRIVE_PWM_RATIO_THOUSANDTHS 560
#define TIMER_FREQ 64000000
#define DEAD_PERIOD 50

//This struct holds the settings for driving
//the H-bridge. Values are calculated in one struct
//then a flag set. When the interrupt sees this flag
//is high is copies over the settings into it's own struct
typedef struct timerSettings
{
    uint32_t timerPeriod;
    int16_t timerPeriodErr; //Err in Hz from target frequency
    int16_t timerPeriodPlusErr; //As above if we use timerPeriod + 1

    uint32_t firstPulsePeriod;
    int32_t firstPulsePeriodErr; //Errors for duty cycle in X/0xFFFF
    int32_t firstPulsePeriodPlusErr;

    uint32_t secondPulsePeriod;
} timerSettings;

static TIM_HandleTypeDef* timer = &htim3;
static timerSettings newSettings;
static volatile bool timerSettingsUpdated = false;

uint16_t pwmDriveSetFrequency(uint16_t driveFrequencyHz,
                              uint16_t pulseWidth)
{
    static uint8_t pwmEnabled = 0;

    uint32_t groundTimePeriod = (TIMER_FREQ * PWM_DRIVE_GROUND_TIME_US) / 
                                (2 * 1000000);

    //
    // Calculate the timer period, plus the errors in Hz for the timer
    // period, plus the error for timer period + 1
    //
    newSettings.timerPeriod = TIMER_FREQ / (driveFrequencyHz * 2);
    newSettings.timerPeriodErr = 
        (TIMER_FREQ / (newSettings.timerPeriod * 2)) - driveFrequencyHz;
    newSettings.timerPeriodPlusErr = 
        (TIMER_FREQ / ((newSettings.timerPeriod + 1) * 2)) - driveFrequencyHz;

    //
    // Calculate settings for first pulse
    //
    int32_t pulse = (newSettings.timerPeriod * 
                      PWM_DRIVE_PWM_RATIO_THOUSANDTHS / 1000) 
                     - groundTimePeriod;

    newSettings.firstPulsePeriod = pulse * pulseWidth;
    newSettings.firstPulsePeriod /= 0xFFFF;

    newSettings.firstPulsePeriodErr = 
        pulseWidth - ((newSettings.firstPulsePeriod * 0xFFFF) / pulse);

    newSettings.firstPulsePeriodPlusErr = 
        pulseWidth - (((newSettings.firstPulsePeriod + 1) * 0xFFFF) / pulse);

    //
    // Calculate settings for second pulse
    //
    newSettings.secondPulsePeriod = 
        newSettings.timerPeriod - (pulse + (groundTimePeriod * 2));

    // Taper to 0 by 75% pulse width
    if(pulseWidth <= 0xBFFD)
    {
        newSettings.secondPulsePeriod = 0;
    }
    else
    {
        newSettings.secondPulsePeriod *= pulseWidth - 0xBFFD;
        newSettings.secondPulsePeriod /=  0xFFFF - 0xBFFD;
    }

    //Modify secondPulsePeriod so it's the point in the counter where
    //the second pulse should turn on, rather than the actual period
    newSettings.secondPulsePeriod = 
        newSettings.timerPeriod - newSettings.secondPulsePeriod;

    timerSettingsUpdated = true;

    if(!pwmEnabled)
    {
        HAL_TIM_Base_Start_IT(timer);
        HAL_TIM_PWM_Start(timer, TIM_CHANNEL_1);
        HAL_TIM_PWM_Start(timer, TIM_CHANNEL_2);
        HAL_TIM_PWM_Start(timer, TIM_CHANNEL_3);
        HAL_TIM_PWM_Start(timer, TIM_CHANNEL_4);
        pwmEnabled = 1;
    }

    return driveFrequencyHz;
}

void pwmDriveDisable()
{
    // Force output off, by setting pulse width to zero
    pwmDriveSetFrequency(21000, 0);  
}

void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef* htim)
{
    static timerSettings settings;
    static int32_t periodErr = 0;
    static int32_t pulseErr = 0;

    //Return if not TIM3
    if(htim->Instance != TIM3)
        return;

    //Copy over settings if updated
    if(timerSettingsUpdated)
    {
        settings = newSettings;
        timerSettingsUpdated = false;
    }

    //Apply dither to timer period, if first pulse period over 500
    //we don't want to apply this dither at very low powers,
    //as it interacts badly with the pulse width adjustment
    if(settings.firstPulsePeriod < 500)
    {
        __HAL_TIM_SET_AUTORELOAD(htim, settings.timerPeriod);
    }
    else if(periodErr > 0)
    {
        __HAL_TIM_SET_AUTORELOAD(htim, settings.timerPeriod + 1);
        periodErr += settings.timerPeriodPlusErr;
    }
    else
    {
        __HAL_TIM_SET_AUTORELOAD(htim, settings.timerPeriod);
        periodErr += settings.timerPeriodErr;
    }

    //Apply dither to first pulse period (Channel A, high side)
    //if period 400 or under
    if(settings.firstPulsePeriod > 400)
    {
        __HAL_TIM_SET_COMPARE(htim, TIM_CHANNEL_2, settings.firstPulsePeriod);
    }
    else if(pulseErr > 0)
    {
        __HAL_TIM_SET_COMPARE(htim, TIM_CHANNEL_2, settings.firstPulsePeriod + 1);
        pulseErr += settings.firstPulsePeriodPlusErr;
    }
    else
    {
        __HAL_TIM_SET_COMPARE(htim, TIM_CHANNEL_2, settings.firstPulsePeriod);
        pulseErr += settings.firstPulsePeriodErr;
    }

    //Channel B high side
    __HAL_TIM_SET_COMPARE(htim, TIM_CHANNEL_1, settings.secondPulsePeriod);

    //Channel A low side
    __HAL_TIM_SET_COMPARE(htim, TIM_CHANNEL_3, 
                          settings.firstPulsePeriod + DEAD_PERIOD);

    //Channel B low side
    __HAL_TIM_SET_COMPARE(htim, TIM_CHANNEL_4, 
                          settings.secondPulsePeriod - DEAD_PERIOD);
}

