/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include "stm32g0xx_hal.h"

#include "driverApplication.h"
#include "driverApplicationSetup.h"

// Common code modules
#include "discPumpCore.h"
#include "settings.h"
#include "uartHandler.h"
#include "controlMethods.h"
#include "sampling.h"
#include "settingsStorage.h"

#include "i2cSensirionFlowSensor.h"
#include "i2cSensirionFlowSensor2.h"

static discPumpCore_handle dpHandle;
static settings* sHandle;

// Functions for handling settings
static void updateUserFacingSettings();
static void updateDiscPumpCoreSettings();
static void checkControlMethod();
static void handlePidControlMethod();
static void handleBangBangControlMethod();

// Helper functions for accessing "sources" which are analog / digital 
// values which update (like an analog input, or pressure sensor reading)
static float (*getSource(settings_sources source))(void);
static float setValSource();
static float analogASource();
static float analogBSource();
static float analogCSource();
static float deadSource();
static float digitalPressureSource();
static float flowSource();

void runApplication()
{
    // Restore settings from flash if any saved
    settingsStorage_RestoreSavedSettings();

    // Init low/mid level hw and modules
    dpHandle = applicationSetup_init();

    // Get a handle for driver settings
    sHandle = settings_getHandle();


    while(1)
    {
        uint32_t startTick = HAL_GetTick();

        // Service underlying code modules
        applicationSetup_tick();  // ~450uS on G0

        // Update any user facing settings 
        updateUserFacingSettings();

        // Check discPumpCoreSettings
        updateDiscPumpCoreSettings();

        // Disable the pump if disabled
        // Otherwise check the control method, and run as appropriate
        if(!sHandle->pumpEnable->value)
            discPumpCore_disable(dpHandle);
        else
            checkControlMethod();  // ~100uS on G0

        while(startTick == HAL_GetTick());
    }
}

static void updateUserFacingSettings()
{
    static uint8_t settingIdx = 0;

    // Round robin updates on each tick, as settings_writeFloat is quite slow
    // TODO - can we speed up writting float settings??
    switch(settingIdx)
    {
        case 0:
            settings_writeFloat(sHandle->measDriveVolts, 
                    sampling_getChannelFiltered(DRIVE_MILLI_VOLTS) / 1000.0);
            break;

        case 1:
            settings_writeFloat(sHandle->measDriveMilliamps, 
                    sampling_getChannelFiltered(DRIVE_MICRO_AMPS) / 1000.0);
            break;

        case 2:
            settings_writeFloat(sHandle->measDriveMilliwatts, 
                    (float)sampling_getChannelFiltered(DRIVE_MICRO_AMPS) *
                    (float)sampling_getChannelFiltered(DRIVE_MILLI_VOLTS)
                    / 1000000.0);
            break;

        case 3:
            settings_writeFloat(sHandle->digitalPressure, 
                                digitalPressureSource());
            break;

        case 4:
            settings_writeFloat(sHandle->measAna3, 
                                analogCSource());
            break;

        case 5:
            if(sHandle->storeCurrentSettings->value != 0)
            {
                // Reset storeCurrentSettings to zero
                settings_writeInt(sHandle->storeCurrentSettings, 0);
                settingsStorage_SaveCurrentSettings();
            }
            break;

        case 6:
            settings_writeInt(sHandle->measDriveFreq, 
                    sampling_getChannelFiltered(ACTUAL_DRIVE_FREQUENCY_HZ));
            break;

        case 7:
            settings_writeFloat(sHandle->measAna1, analogASource());
            break;

        case 8:
            settings_writeFloat(sHandle->measAna2, analogBSource());
            break;
        case 9:
        	settings_writeFloat(sHandle->measFlow, flowSource());
        	break;

        default:
            break;
    }

    if(++settingIdx > 9)
        settingIdx = 0;

    // Enable stream mode if requested
    if(sHandle->enableStreamMode->updated)
    {
        settings_updateNotedInt(sHandle->enableStreamMode);
        uartHandler_setStreaming(sHandle->enableStreamMode->value == 1);
    }
}

static void updateDiscPumpCoreSettings()
{
    if(sHandle->powerLimitMilliwatts->updated)
    {
        settings_updateNotedInt(sHandle->powerLimitMilliwatts);

        discPumpCore_setPowerLimit(dpHandle,
                                   sHandle->powerLimitMilliwatts->value);
    }

    if(sHandle->useFrequencyTracking->updated ||
       sHandle->manualDriveFrequency->updated)
    {
        settings_updateNotedInt(sHandle->useFrequencyTracking);
        settings_updateNotedInt(sHandle->manualDriveFrequency);

        if(sHandle->useFrequencyTracking->value)
        {
            discPumpCore_enableFrequencyTracking(dpHandle);
        }
        else
        {
            discPumpCore_setManualFrequency(dpHandle, 
                    sHandle->manualDriveFrequency->value);
        }
    }
}

static void checkControlMethod()
{
    // Check the current control mode, and update accordingly
    switch(sHandle->mode->value)
    {
        case SETTINGS_MODE_MANUAL:
            discPumpCore_setPower(dpHandle, 
                    (*getSource(sHandle->manualModeSource->value))() * 1000);
            break;

        case SETTINGS_MODE_PID:
            // Checks settings, and updates the PID config is necessary
            handlePidControlMethod();

            // Limit PID output to positive values
            float output = controlMethods_tick();
            if(output < 0) output = 0;

            // Run the control loop, using power in mW as the target
            discPumpCore_setPower(dpHandle, output * 1000.0);
            break;

        case SETTINGS_MODE_BANGBANG:
            // Checks settings, and updates the bangBang config is necessary
            handleBangBangControlMethod();

            // Run the control loop, using power in mW as the target
            discPumpCore_setPower(dpHandle, controlMethods_tick() * 1000);
            break;
    }
}

static void handlePidControlMethod()
{
    if(sHandle->pumpEnable->updated             ||
       sHandle->mode->updated                   ||
       sHandle->pidModeSetpointSource->updated  ||
       sHandle->pidModeMeasSource->updated      ||
       sHandle->pidProportionalCoeff->updated   ||
       sHandle->pidIntegralCoeff->updated       ||
       sHandle->pidIntegralLimitCoeff->updated  ||
       sHandle->pidDifferentialCoeff->updated)
    {
        // Reset loop if control mode or integral settings have changed
        // or pump just turned on and resetPidOnTurnon is enabled
        bool resetLoop = (  sHandle->mode->updated ||
                            sHandle->pidIntegralCoeff->updated ||
                            sHandle->pidIntegralLimitCoeff->updated
                         ) 
                        ||
                         (
                            sHandle->pumpEnable->updated &&
                            sHandle->resetPidOnTurnon->value
                         );

        //Make sure all the updated settings have been notified
        settings_updateNotedInt(sHandle->pumpEnable);
        settings_updateNotedInt(sHandle->mode);
        settings_updateNotedInt(sHandle->pidModeSetpointSource);
        settings_updateNotedInt(sHandle->pidModeMeasSource);
        settings_updateNotedFloat(sHandle->pidProportionalCoeff);
        settings_updateNotedFloat(sHandle->pidIntegralCoeff);
        settings_updateNotedFloat(sHandle->pidIntegralLimitCoeff);
        settings_updateNotedFloat(sHandle->pidDifferentialCoeff);

        // Update the PID loop settings
        controlMethods_pidSettings pidSettings;

        pidSettings.getInputValue = 
            getSource(sHandle->pidModeMeasSource->value);

        pidSettings.getSetPoint = 
            getSource(sHandle->pidModeSetpointSource->value);

        pidSettings.proportionalTerm = sHandle->pidProportionalCoeff->value;
        pidSettings.integralTerm = sHandle->pidIntegralCoeff->value;
        pidSettings.integralMin = 0;
        pidSettings.integralMax = sHandle->pidIntegralLimitCoeff->value;
        pidSettings.differentialTerm = sHandle->pidDifferentialCoeff->value;
        
        controlMethods_startPID(pidSettings, resetLoop);
    }
}

static void handleBangBangControlMethod()
{
    if(sHandle->pumpEnable->updated                     ||
       sHandle->mode->updated                           ||
       sHandle->bangBangMeasSource->updated             ||
       sHandle->bangBangLowerThresh->updated            ||
       sHandle->bangBangUpperThresh->updated            ||
       sHandle->bangBangLowerPowerMilliwatts->updated   ||
       sHandle->bangBangUpperPowerMilliwatts->updated)
    {
        //Make sure all the updated settings have been notified
        settings_updateNotedInt(sHandle->pumpEnable);
        settings_updateNotedInt(sHandle->mode);
        settings_updateNotedInt(sHandle->bangBangMeasSource);
        settings_updateNotedFloat(sHandle->bangBangLowerThresh);
        settings_updateNotedFloat(sHandle->bangBangUpperThresh);
        settings_updateNotedFloat(sHandle->bangBangLowerPowerMilliwatts);
        settings_updateNotedFloat(sHandle->bangBangUpperPowerMilliwatts);

        controlMethods_bangBangSettings bbSettings;

        bbSettings.getInputValue = 
            getSource(sHandle->bangBangMeasSource->value);

        bbSettings.lowerThresh = sHandle->bangBangLowerThresh->value;
        bbSettings.upperThresh = sHandle->bangBangUpperThresh->value;

        bbSettings.lowerControlValue = 
            sHandle->bangBangLowerPowerMilliwatts->value;

        bbSettings.upperControlValue = 
            sHandle->bangBangUpperPowerMilliwatts->value;

        controlMethods_startBangBang(bbSettings);
    }
}

static float (*getSource(settings_sources source))(void)
{
    switch(source)
    {
        case SETTINGS_SOURCE_SETVAL:
            return &setValSource;

        case SETTINGS_SOURCE_ANA1:
            return &analogASource;

        case SETTINGS_SOURCE_ANA2:
            return &analogBSource;

        case SETTINGS_SOURCE_ANA3:
            return &analogCSource;

        case SETTINGS_SOURCE_PRESSURE:
            return &digitalPressureSource;

        case SETTINGS_SOURCE_FLOW:
            return &flowSource;

        default:
            return &deadSource;
    }
}

static float setValSource()
{
    return sHandle->setVal->value;
}

static float analogASource()
{
    float val = 0;

    //Get raw analog value, and convert to 0.0->1.0
    val = sampling_getChannelFiltered(ANA_A) / 32768.0;

    //Apply offset and gain
    val *= sHandle->ana1Gain->value;
    val += sHandle->ana1Offset->value;

    return val;
}

static float analogBSource()
{
    float val = 0;

    //Get raw analog value, and convert to 0.0->1.0
    val = sampling_getChannelFiltered(ANA_B) / 32768.0;

    //Apply offset and gain
    val *= sHandle->ana2Gain->value;
    val += sHandle->ana2Offset->value;

    return val;
}

static float analogCSource()
{
    //Get raw analog value, and convert to 0.0->1.0
    float val = sampling_getChannelFiltered(ANA_C) / 32768.0;

    //Apply offset and gain
    val *= sHandle->ana3Gain->value;
    val += sHandle->ana3Offset->value;

    return val;
}

static float deadSource()
{
    return 0; 
}


static float pressure_scalings [] =
{1.0, 				// 0 = mbar	= default
0.750062, 			// 1 = mmHg
0.0145038,			// 2 = PSI
0.1,				// 3 = kPa
0.02953, 			// 4 = inHg
0.401865,			// 5 = inH2O
1.01972};			// 6 = cmH2O

static float digitalPressureSource()
{
#ifndef GAUGE_PRESSURE
    float pressuremBar = (sampling_getChannelFiltered(HONEYWELL_MPR_UBAR) -
                          sampling_getChannelFiltered(BAROMETER_UBAR)) / 1000.0;
#else
    float pressuremBar = sampling_getChannelFiltered(HONEYWELL_MPR_UBAR) / 1000.0;
#endif

    float pressure = pressuremBar * pressure_scalings[sHandle->DigitalPressureMeasUnit->value];
    pressure += sHandle->digitalPressureOffset->value;

    return pressure;
}

static float flow_scalings [] =
{0.000000001, 	// 0 = L/min
0.000001,		// 1 = mL/min = default
0.001, 			// 2 = uL/min
1.0};			// 3 = nL/min

static float flowSource()
{
	//data stored in nL/min  reminiscent of the the sampling class only takes integers. We want to output in ml/min by def

	// this is a hacky way of getting the flow - we really should be using the sampling.c (sampling_getChannelFiltered)
	// like the rest. Problem is sampling only works with 32bit ints which is not big enough for the range of values expected
	// TODO think if we can expand the sampling method to include float / parse floats
	float flow_nLmin =	flowSensor_getFlow_nLperMin() + flowSensor2_getFlow_nLperMin();
	return flow_nLmin * flow_scalings[sHandle->FlowMeasUnit->value];
}

