/*
 * Copyright (C) LEE Ventus - All Rights Reserved
 * Unauthorised copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */

#include "driverApplicationSetup.h"

// Common code modules
#include "sampling.h"
#include "uartHandler.h"
#include "boostController.h"
#include "i2cMasterHandler.h"
#include "i2cHoneywellMPR.h"
#include "i2cMs563702BA02Handler.h"
#include "settings.h"
#include "discPumpCore.h"


// HAL modules
#include "stm32g0xx_hal.h"
#include "adc.h"
#include "tim.h"
#include "i2c.h"

//
// Driver specific modules
//
#include "usart.h"

#include "pwmDrive.h"

#include "GPIO_Handler.h"
#include "i2cSensirionFlowSensor.h"
#include "i2cSensirionFlowSensor2.h"
#include "i2cTlc59116GaugeHandler.h"
#include "i2cLP5861GaugeHandler.h"

static uint16_t driveFreq;

/******************************************************************************
 * Sampling setup
 ******************************************************************************/
static uint16_t adcBuf[ANALOG_CHANNEL_COUNT*2];
static int16_t currentOffset = 0;

static inline void processAdcSamples(uint16_t* buf)
{
    sampling_updateChannel(DRIVE_MICRO_AMPS, 
                           (int16_t)buf[0] - currentOffset);
    sampling_updateChannel(ANA_A, buf[1]);
    sampling_updateChannel(ANA_B, buf[2]);
    sampling_updateChannel(ANA_C, buf[3]);
    sampling_updateChannel(DRIVE_MILLI_VOLTS, buf[4]);
}

// HAL callbacks when a scan of the analog channels is complete
void HAL_ADC_ConvCpltCallback(ADC_HandleTypeDef* hadc)
{
    processAdcSamples(adcBuf + ANALOG_CHANNEL_COUNT);
}

void HAL_ADC_ConvHalfCpltCallback(ADC_HandleTypeDef* hadc)
{
    processAdcSamples(adcBuf);
}

static void setupSamplingModule()
{
    sampling_channelConfig config[SAMPLING_CHANNEL_COUNT];

    // 4.26x multiplication factor
    config[DRIVE_MICRO_AMPS].oversamplingFactor = 2;
    config[DRIVE_MICRO_AMPS].scalingMulti = 17460;
    config[DRIVE_MICRO_AMPS].scalingRightShift = 12;

    // 2.019x multiplication factor
    config[DRIVE_MILLI_VOLTS].oversamplingFactor = 2;
    config[DRIVE_MILLI_VOLTS].scalingMulti = 16540;
    config[DRIVE_MILLI_VOLTS].scalingRightShift = 13;

    // All analog input channels 
    config[ANA_A].oversamplingFactor = 6;
    config[ANA_A].scalingMulti = 1;
    config[ANA_A].scalingRightShift = 0;

    config[ANA_B].oversamplingFactor = 6;
    config[ANA_B].scalingMulti = 1;
    config[ANA_B].scalingRightShift = 0;


    config[ANA_C].oversamplingFactor = 6;
    config[ANA_C].scalingMulti = 1;
    config[ANA_C].scalingRightShift = 0;

    config[ACTUAL_DRIVE_FREQUENCY_HZ].oversamplingFactor = 4;
    config[ACTUAL_DRIVE_FREQUENCY_HZ].scalingMulti = 1;
    config[ACTUAL_DRIVE_FREQUENCY_HZ].scalingRightShift = 0;

    config[HONEYWELL_MPR_UBAR].oversamplingFactor = 0;
    config[HONEYWELL_MPR_UBAR].scalingMulti = HONEYWELL_MPR_UBAR_SCALING;
    config[HONEYWELL_MPR_UBAR].scalingRightShift = HONEYWELL_MPR_UBAR_SHIFT;

    // Just store the barometric pressure
    config[BAROMETER_UBAR].oversamplingFactor = 0;
    config[BAROMETER_UBAR].scalingMulti = 1;
    config[BAROMETER_UBAR].scalingRightShift = 0;


    // Just store the flow rate
	config[SENSIRION_FLOW_SENSOR].oversamplingFactor = 0;
	config[SENSIRION_FLOW_SENSOR].scalingMulti = 1;
	config[SENSIRION_FLOW_SENSOR].scalingRightShift = 0;

    sampling_setup(SAMPLING_CHANNEL_COUNT, config);

    // Perform ADC calibration, and start running
    HAL_ADCEx_Calibration_Start(&hadc1);
    HAL_ADC_Start_DMA(&hadc1, 
                      (uint32_t*)adcBuf, 
                      sizeof(adcBuf)/2);

    // Run the ADC 250x
    for(uint8_t count=0; count<250; count++)
    {
        // Trigger the ADC, and wait till it's complete
        hadc1.Instance->CR |= ADC_CR_ADSTART;  
        while(hadc1.Instance->CR & ADC_CR_ADSTART);
    }

    // Divide by 4.26x, as the raw ADC value is scaled up by 4.26x, and we
    // apply the offset to the raw ADC value
    currentOffset = sampling_getChannelFiltered(DRIVE_MICRO_AMPS) / 4.26;
}

static void tickSamplingModule()
{
    // Update pressure from Honeywell MPR
    int32_t counts = honeywellMPR_getCounts();
    counts -= HONEYWELL_MPR_UBAR_OFFSET;  // Subtract offset from returned 24-bit value. This is usually 10%
    counts >>= 8;  // Convert down to 16-bit value
    sampling_updateChannel(HONEYWELL_MPR_UBAR, counts);

#ifndef GAUGE_PRESSURE
    // Get latest pressure from the barometer
    sampling_updateChannel(BAROMETER_UBAR, ms563702ba03_getPressureuBar());
#endif

    // Get latest pressure from the flow sensor
    //data stored in nL/min as the sampling class only takes integers

    //sampling_updateChannel(SENSIRION_FLOW_SENSOR, ((int32_t) flowSensor_getFlow_nLperMin()) + ((int32_t) flowSensor2_getFlow_nLperMin()));

}

/******************************************************************************
 * GPIO setup
 ******************************************************************************/
gpio_instance GPIOA_instance;
gpio_instance GPIOB_instance;
gpio_instance GPIOC_instance;
gpio_instance GPIOD_instance;

#define GPIO_A_PIN				GPIO_PIN_0
#define GPIO_A_PERIPHERAL		GPIOD

#define GPIO_B_PIN				GPIO_PIN_1
#define GPIO_B_PERIPHERAL		GPIOD
#define GPIO_B_TIMER_CHANNEL 	TIM_CHANNEL_1
#define GPIO_B_TIMER_COUNTER	(&htim17)
#define GPIO_B_TIMER_ALTERNATEF GPIO_AF2_TIM17

#define GPIO_C_PIN				GPIO_PIN_2
#define GPIO_C_PERIPHERAL		GPIOD

#define GPIO_D_PIN				GPIO_PIN_3
#define GPIO_D_PERIPHERAL		GPIOD

void setGpioAstate( bool state) { HAL_GPIO_WritePin(GPIO_A_PERIPHERAL, GPIO_A_PIN, state); }
void setGpioBstate( bool state) { HAL_GPIO_WritePin(GPIO_B_PERIPHERAL, GPIO_B_PIN, state); }
void setGpioCstate( bool state) { HAL_GPIO_WritePin(GPIO_C_PERIPHERAL, GPIO_C_PIN, state); }
void setGpioDstate( bool state) { HAL_GPIO_WritePin(GPIO_D_PERIPHERAL, GPIO_D_PIN, state); }

bool getGpioAState() { return HAL_GPIO_ReadPin(GPIO_A_PERIPHERAL, GPIO_A_PIN); }
bool getGpioBState() { return HAL_GPIO_ReadPin(GPIO_B_PERIPHERAL, GPIO_B_PIN); }
bool getGpioCState() { return HAL_GPIO_ReadPin(GPIO_C_PERIPHERAL, GPIO_C_PIN); }
bool getGpioDState() { return HAL_GPIO_ReadPin(GPIO_D_PERIPHERAL, GPIO_D_PIN); }

void startGpioBTimPulse (uint8_t num_pulses, uint16_t period, uint16_t duration)
{
	// it is better to use the Timer counter strucutres and Hal functions to set values
	// however that would be a bit slow (we need to reinitialise the whole thing) so doing this with direct register edit

	GPIO_B_TIMER_COUNTER->Instance->RCR = num_pulses; // this is an 8bit value so need to cap the register to 0 to 250
	GPIO_B_TIMER_COUNTER->Instance->ARR = period;
	GPIO_B_TIMER_COUNTER->Instance->CCR1 = duration;
	/* Generate an update event to reload the the repetition counter value immediately */
	GPIO_B_TIMER_COUNTER->Instance->EGR |= TIM_EGR_UG;

	// set the counter value to the pulse duration. This is dont to front load the pulse as this is the behaviour we want
	// This is necessary as the timer only supports upcounting and the "pulse" is on only when the counter CNT > duration CCR1
	// has to be duration-1 and not duration otherwise a ghost 1count (=10us) pulse is generated for some reason.
	// This does cause a 1 count (=10us) delay to the start of the pulse but this is acceptable given we can take up to 1ms to process the command.
	GPIO_B_TIMER_COUNTER->Instance->CNT = duration-1;	// duration has to be >= 1

	HAL_TIM_OnePulse_Start(GPIO_B_TIMER_COUNTER, GPIO_B_TIMER_CHANNEL);

	// for some reason had to stop it first. Clears the state maybe
	HAL_TIM_Base_Stop(GPIO_B_TIMER_COUNTER);
	HAL_TIM_Base_Start(GPIO_B_TIMER_COUNTER);

}

void setGpioBTimControlMode(uint32_t mode)
{
	// use one of TIM_OCMODE_PWM1, TIM_OCMODE_PWM2, TIM_OCMODE_FORCED_ACTIVE, TIM_OCMODE_FORCED_INACTIVE
	uint32_t CCMR1_Reg_temp = 0;
	CCMR1_Reg_temp = GPIO_B_TIMER_COUNTER->Instance->CCMR1;
	CCMR1_Reg_temp &= ~TIM_CCMR1_OC1M_Msk; 	//zero the relevant bits
	CCMR1_Reg_temp |= mode; 		//set the bits for whether the channel is inverted or not
	GPIO_B_TIMER_COUNTER->Instance->CCMR1 = CCMR1_Reg_temp;

}

static void setupGPIOModule(GPIO_TypeDef *GPIOx, uint16_t GPIO_Pin, GPIO_mode setup_pin_mode, gpio_instance* pin, TIM_HandleTypeDef* timer_counter, uint32_t GPIO_ALTERNATIVE_timer_channel)
{

	GPIO_InitTypeDef GPIO_InitStruct = {0};
	GPIO_InitStruct.Pin = GPIO_Pin;
	bool def_pin_state = false;

	switch (setup_pin_mode){
		case OUTPUT_HIGH_SPEED_DEF_ON:
			// By default the GPIO is set to input with no pull up or pull down (essentially high Z)
			// need to configure the GPIO to the timer counter

			//the Timer counter is already configured so not much to alter except for the the default state of the channel
			// important to configure the timer counter to have zero in the period, duration and repetition counter to prevent the first pulse from messing up
			// need to set the default state before we initialise the GPIO
			setGpioBTimControlMode(TIM_OCMODE_PWM1);
			pin->startTimPulse(0,0,1);	//make the change of the default state take place

			//configure the pin
			GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
			GPIO_InitStruct.Pull = GPIO_NOPULL;
			GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
			GPIO_InitStruct.Alternate = GPIO_ALTERNATIVE_timer_channel;

			def_pin_state = GPIO_PIN_SET;
		break;

		case OUTPUT_HIGH_SPEED_DEF_OFF:
			// By default the GPIO is set to input with no pull up or pull down (essentially high Z)
			// need to configure the GPIO to the timer counter

			//the Timer counter is already configured so not much to alter except for the the default state of the channel
			// important to configure the timer counter to have zero in the period, duration and repetition counter to prevent the first pulse from messing up
			// need to set the default state before we initialise the GPIO
			setGpioBTimControlMode(TIM_OCMODE_PWM2);
			pin->startTimPulse(0,0,1);	//make the change of the default state take place

			//configure the pin
			GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
			GPIO_InitStruct.Pull = GPIO_NOPULL;
			GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
			GPIO_InitStruct.Alternate = GPIO_ALTERNATIVE_timer_channel;

			def_pin_state = GPIO_PIN_RESET;
			break;

		case OUTPUT_LOW_SPEED_DEF_ON:
			//configure the pin

			GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
			GPIO_InitStruct.Pull = GPIO_NOPULL;
			GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;

			def_pin_state = GPIO_PIN_SET;
			break;

		case OUTPUT_LOW_SPEED_DEF_OFF:
			//configure the pin
			GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
			GPIO_InitStruct.Pull = GPIO_NOPULL;
			GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;

			def_pin_state = GPIO_PIN_RESET;
			break;

		case INPUT_NO_PULL:
			//configure the pin
			GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
			GPIO_InitStruct.Pull = GPIO_NOPULL;
			break;

		case INPUT_PULL_UP:
			//configure the pin
			GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
			GPIO_InitStruct.Pull = GPIO_PULLUP;
			break;

		case INPUT_PULL_DOWN:
			//configure the pin
			GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
			GPIO_InitStruct.Pull = GPIO_PULLDOWN;
			break;

		case DISABLED:
			GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
			GPIO_InitStruct.Pull = GPIO_NOPULL;

		default:
			// the code should never be here but setting it as disabled
			GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
			GPIO_InitStruct.Pull = GPIO_NOPULL;
			setup_pin_mode = DISABLED;
	}


	HAL_GPIO_Init(GPIOx, &GPIO_InitStruct);

	GPIO_Handler_init(pin, setup_pin_mode, def_pin_state);
}

static void GPIO_init_all()
{
	GPIO_Handler_set_TC_states(TIM_OCMODE_PWM1, TIM_OCMODE_PWM2, TIM_OCMODE_FORCED_ACTIVE, TIM_OCMODE_FORCED_INACTIVE);

	GPIOA_instance.pin_state_setting			= settings_getHandle()->pinStateGPIOA;
	GPIOA_instance.pin_pulse_duration_setting	= settings_getHandle()->pinPulseDurationGPIOA;
	GPIOA_instance.pin_pulse_period_setting		= settings_getHandle()->pinPulsePeriodGPIOA;
	GPIOA_instance.set_Output_GPIO				= &setGpioAstate;
	GPIOA_instance.get_Input_GPIO				= &getGpioAState;
	setupGPIOModule(GPIO_A_PERIPHERAL, GPIO_A_PIN, settings_getHandle()->pinModeGPIOA->value, &GPIOA_instance, NULL, 0);

	GPIOB_instance.pin_state_setting			= settings_getHandle()->pinStateGPIOB;
	GPIOB_instance.pin_pulse_duration_setting	= settings_getHandle()->pinPulseDurationGPIOB;
	GPIOB_instance.pin_pulse_period_setting		= settings_getHandle()->pinPulsePeriodGPIOB;
	GPIOB_instance.set_Output_GPIO				= &setGpioBstate;
	GPIOB_instance.get_Input_GPIO				= &getGpioBState;
	GPIOB_instance.startTimPulse				= &startGpioBTimPulse;
	GPIOB_instance.setTimControlMode			= &setGpioBTimControlMode;
	setupGPIOModule(GPIO_B_PERIPHERAL, GPIO_B_PIN, settings_getHandle()->pinModeGPIOB->value, &GPIOB_instance, GPIO_B_TIMER_COUNTER, GPIO_B_TIMER_ALTERNATEF);

	GPIOC_instance.pin_state_setting			= settings_getHandle()->pinStateGPIOC;
	GPIOC_instance.pin_pulse_duration_setting	= settings_getHandle()->pinPulseDurationGPIOC;
	GPIOC_instance.pin_pulse_period_setting		= settings_getHandle()->pinPulsePeriodGPIOC;
	GPIOC_instance.set_Output_GPIO				= &setGpioCstate;
	GPIOC_instance.get_Input_GPIO				= &getGpioCState;
	setupGPIOModule(GPIO_C_PERIPHERAL, GPIO_C_PIN, settings_getHandle()->pinModeGPIOC->value, &GPIOC_instance, NULL, 0);

	GPIOD_instance.pin_state_setting			= settings_getHandle()->pinStateGPIOD;
	GPIOD_instance.set_Output_GPIO				= &setGpioDstate;
	GPIOD_instance.get_Input_GPIO				= &getGpioDState;
	setupGPIOModule(GPIO_D_PERIPHERAL, GPIO_D_PIN, INPUT_PULL_UP, &GPIOD_instance, NULL, 0);
	//we do not allow GPIO_D to be configured as it's connected to the button on the Eval kit
	//if someone configures it to otuput they can burn the pin out
}

static void GPIO_tick_all()
{
	GPIO_Handler_tick(&GPIOA_instance);
	GPIO_Handler_tick(&GPIOB_instance);
	GPIO_Handler_tick(&GPIOC_instance);
	GPIO_Handler_tick(&GPIOD_instance);
}


/******************************************************************************
 * UART Setup
 * Includes integration with settings.c/h
 ******************************************************************************/

#define ANALOG_PRESSURE 8
#define DIGITAL_PRESSURE 39

static uint8_t rxDmaTail = 0;
static volatile uint8_t uartRxBuffer[256];
static uint8_t streamSettings[] = 
{
    0,  // Pump enabled
    3,  // Voltage 
    4,  // Current
    6,  // Drive frequency
    7,  // Analog A

	//keep the pressure setting in 6th place as that is used for auto detecting the i2c pressure sensor
	// value gets overwritten if digital pressure sensor is detected (on GP driver)
	ANALOG_PRESSURE,  // Analog B
    9,  // Analog C 
    32  // Sensirion I2C flow sensor
}; 

void HAL_UART_ErrorCallback(UART_HandleTypeDef *huart)
{
    //Error is usually on receive, if someone sent us garbage.
    //Restart the receive
    rxDmaTail = 0;
    HAL_UART_Receive_DMA(&huart1, (uint8_t*)&uartRxBuffer, 256);
}

static bool getUnreadBytes(uint8_t* unreadBytes, uint8_t* len)
{
    uint8_t rxDmaHead = 256 - huart1.hdmarx->Instance->CNDTR;
    uint8_t count = 0;

    // Just return false if no bytes received
    if(rxDmaTail == rxDmaHead)
        return false;

    // Otherwise copy the bytes over to the supplied buffer
    while(rxDmaTail != rxDmaHead)
    {
        unreadBytes[count++] = uartRxBuffer[rxDmaTail++];
    }

    // Copy over the number of bytes
    *len = count;

    return true;  // Indicate some bytes have been received
}

static bool sendBytes(uint8_t* bytesToWrite, uint16_t len)
{
    return HAL_UART_Transmit_DMA(&huart1, bytesToWrite, len) == HAL_OK;
}

static void setupUartHandlerModule()
{
    // Uart Handler setup config
    uartHandler_setup uartSetup;
    uartSetup.getSetting = settings_getSettingAsString;
    uartSetup.setSetting = settings_setSettingAsString;
    uartSetup.settingsToStream = streamSettings;
    uartSetup.settingsToStreamCount = 8;
    uartSetup.ticksPerStream = 16;
    uartSetup.getUnreadBytes = getUnreadBytes;
    uartSetup.sendBytes = sendBytes;

    uartHandler_init(uartSetup);
    uartHandler_setStreaming(false);

    // Start receiving bytes
    HAL_UART_Receive_DMA(&huart1, (uint8_t*)&uartRxBuffer, 256);
}

/******************************************************************************
 * Slave I2C Setup
 ******************************************************************************/

/******************************************************************************
 * Master I2C setup
 ******************************************************************************/
static volatile i2c_state masterI2cState = I2C_IDLE;

static I2C_HandleTypeDef* i2cMasterHandle = &hi2c1;

void HAL_I2C_MasterTxCpltCallback(I2C_HandleTypeDef *hi2c)
{
    masterI2cState = I2C_SUCCESS;
}

void HAL_I2C_MasterRxCpltCallback(I2C_HandleTypeDef *hi2c)
{
    masterI2cState = I2C_SUCCESS;
}

void HAL_I2C_ErrorCallback(I2C_HandleTypeDef *hi2c)
{
    // TODO - determine why the slave I2C bus raises lots of errors, when
    // it appears to be working normally.
    // Is it just what happens when the master NACKs the last byte in a receive
    // sequence?
    if(hi2c == i2cMasterHandle)
        masterI2cState = I2C_FAILURE;
}

static bool checkDeviceIsAvailable(uint8_t addr)
{
    return HAL_OK == HAL_I2C_IsDeviceReady(i2cMasterHandle, addr << 1, 100, 10);
}

static void masterSendBytes(uint8_t addr, uint8_t* buf, uint8_t bytesToSend)
{
    HAL_I2C_Master_Transmit_IT(i2cMasterHandle, addr << 1, buf, bytesToSend);
    masterI2cState = I2C_BUSY;
}

static void masterReadBytes(uint8_t addr, uint8_t* buf, uint8_t bytesToRead)
{
    HAL_I2C_Master_Receive_IT(i2cMasterHandle, addr << 1, buf, bytesToRead);
    masterI2cState = I2C_BUSY;
}

#define INDX_HONEYWELL_PSENSOR 0
#define INDX_LP5861_DEVICE 6
#define INDX_LP5861_BRIGHTNESS 7

static void setupMasterI2CHandler()
{
    // Get I2C devices
	i2c_device devices[8];
	bool isDeviceActive[8];



    devices[INDX_HONEYWELL_PSENSOR] = honeywellMPR_getDevice();
    devices[1] = ms563702ba03_getDevice();
    devices[2] = flowSensor_getDevice();
    devices[3] = flowSensor2_getDevice();
    devices[4] = flowSensor_air_getDevice();

    // motherboard LED drivers
    devices[5] = tlc59116_getDevice();
    devices[INDX_LP5861_DEVICE] = LP5861_getDevice();
	devices[INDX_LP5861_BRIGHTNESS] = LP5861BrigthnessControl_getDevice();
	//the above are set to the last two indexes 6, 7. They are macroed as the indexes are important for the end of this funcion
	//must be kept as the last two devices in the order shown - device first and then brightness

    // Setup for i2cMasterhandler
    i2c_setup setup;
    setup.checkDeviceIsAvailable = checkDeviceIsAvailable;
    setup.sendBytes = masterSendBytes;
    setup.readBytes = masterReadBytes;
    setup.devices = devices;
    setup.deviceCount = INDX_LP5861_DEVICE + 1; //dont include the brihtness one as we will force add it later

    i2cMaster_init(setup, isDeviceActive);

    if(isDeviceActive[INDX_LP5861_DEVICE])
    {
    	// if the LED chip is detected add the britness control
    	// as the chip does not respond to the check call as it has an internal Enable register set to off
    	i2cMaster_forceAddDevice(devices[INDX_LP5861_BRIGHTNESS], LP5861_BRIGHTNESS_INITIAL_DELAY);
    }

    if(isDeviceActive[INDX_HONEYWELL_PSENSOR])
    {
    	//if the digital pressure sensor is detected, report the digital pressure, not analog
    	streamSettings[5] = DIGITAL_PRESSURE;
    }
}

static void tickMasterI2CHandler()
{
    i2cMaster_tick(masterI2cState);

    if(masterI2cState != I2C_BUSY)
        masterI2cState = I2C_IDLE;
}

/******************************************************************************
 * Boost Controller Setup
 ******************************************************************************/

#define BOOST_NUM_SAMPLES (8)  // 31.25kHz update rate
static TIM_HandleTypeDef* boostTimer = &htim1;

static uint32_t boostPwmBuf[BOOST_NUM_SAMPLES * 2];

// HAL callbacks when one half of boostPwmBuf is consumed
void HAL_TIM_PWM_PulseFinishedCallback(TIM_HandleTypeDef *htim)
{
    hadc1.Instance->CR |= ADC_CR_ADSTART;
    boostController_getNextSamples(sampling_getChannelRaw(DRIVE_MILLI_VOLTS), 
                                   boostPwmBuf + BOOST_NUM_SAMPLES, 
                                   BOOST_NUM_SAMPLES);
}

void HAL_TIM_PWM_PulseFinishedHalfCpltCallback(TIM_HandleTypeDef *htim)
{
    hadc1.Instance->CR |= ADC_CR_ADSTART;
    boostController_getNextSamples(sampling_getChannelRaw(DRIVE_MILLI_VOLTS), 
                                   boostPwmBuf, BOOST_NUM_SAMPLES);
}

static void setupBoostControllerModule()
{
    uint16_t const boostPidPTerm = 500;
    uint16_t const boostPidITerm = 50;

    // Setup the Boost Controller
    boostController_init(sampling_getChannelRaw(DRIVE_MILLI_VOLTS),
                         58000, 
                         boostPidPTerm, 
                         boostPidITerm);

    // Init at 0
    boostController_setTarget(0);

            
    // Start the PWM output for the boost controller
    HAL_TIM_PWM_Start_DMA(boostTimer, 
                          TIM_CHANNEL_1, 
                          (uint32_t*) boostPwmBuf, 
                          sizeof(boostPwmBuf)/4);
    HAL_TIMEx_PWMN_Start(boostTimer, TIM_CHANNEL_1);
}

static void emergencyStopBoostConverter()
{
    HAL_TIM_PWM_Stop_DMA(boostTimer, TIM_CHANNEL_1);
}


/******************************************************************************
 * Disc Pump Core setup
 ******************************************************************************/
static discPumpCore_handle handle;
static uint16_t dutyCycle;
static bool pumpOn;

// Helper functions for discPumpCore
static void discPumpSetDriveTarget(uint16_t targetDriveFrequencyHz, 
                                   uint32_t targetDriveVoltageMillivolts,
                                   uint16_t targetDutyCycle)
{
    boostController_setTarget(targetDriveVoltageMillivolts);

    //Set the drive frequency
    // -Use sineWaveGen module for the soft driver
    // -Use other drivers have their own code for setting the driver frequency
    driveFreq = pwmDriveSetFrequency(targetDriveFrequencyHz,
                                     targetDutyCycle);

    //Copy over duty cycle to local variable
    dutyCycle = targetDutyCycle;

    // Update sampling module with the current drive frequency
    sampling_updateChannel(ACTUAL_DRIVE_FREQUENCY_HZ, driveFreq);

    pumpOn = true;
}

//Function pointer to function which should disable the 
//disc pump drive system
static void disableDiscPumpDrive(void)
{
    pwmDriveDisable();

    boostController_setTarget(0);
    pumpOn = false;
}

static discPumpCore_handle setupDiscPumpCoreModule()
{
    //Setup Disc Pump Core
    discPumpCore_setup discPumpCoreSetup;
    discPumpCoreSetup.setDrive = &discPumpSetDriveTarget;
    discPumpCoreSetup.disableDrive = &disableDiscPumpDrive;

    discPumpCoreSetup.useSoftDriveProtocol = false;
    discPumpCoreSetup.maxDriveMillivolts = 50000;

    discPumpCoreSetup.driveFreqTrackingHysteresisHz = 12;
    discPumpCoreSetup.driveFreqTrackingStepSizeHz = 2;

    discPumpCoreSetup.minDriveMillivolts = 5500;
    discPumpCoreSetup.minDutyCycle = 0;
    discPumpCoreSetup.maxDriveRampRateMillivolts = 10000;  // Limited in boost 
    discPumpCoreSetup.driveVoltageOpenLoop = true;

    discPumpCoreSetup.minDriveFreq = 19500;
    discPumpCoreSetup.maxDriveFreq = 23000; 

    return discPumpCore_init(discPumpCoreSetup);
}

/******************************************************************************
 * Low level application setup
 * Groups all the other setup functions together
 ******************************************************************************/
static bool useUart;

discPumpCore_handle applicationSetup_init()
{
    setupSamplingModule();
    setupMasterI2CHandler();
    setupBoostControllerModule();


    GPIO_init_all();

    useUart = true;

    if(useUart)
        setupUartHandlerModule();

    handle = setupDiscPumpCoreModule();

    return handle;
}

void applicationSetup_tick()
{
    GPIO_tick_all();

    // Tick middleware
    if(useUart)
        uartHandler_tick();  

    tickMasterI2CHandler();
    tickSamplingModule();

    int32_t current_ua = sampling_getChannelFiltered(DRIVE_MICRO_AMPS);
    int32_t driveMv = sampling_getChannelFiltered(DRIVE_MILLI_VOLTS);

    discPumpCore_update update;
    update.measuredDriveMillivolts = driveMv;
    update.measuredDriveMicroamps = current_ua > 0 ? current_ua : 0;
    update.actualDriveFreqHz = driveFreq;
    update.actualDutyCycle = dutyCycle;


    discPumpCore_status status = discPumpCore_tick(handle, update);

    tlc59116_setErrorLed(status != DISC_PUMP_CORE_STATUS_OKAY);


    if(status != DISC_PUMP_CORE_STATUS_OKAY)
    	LP5861_led_setState(RGB_LED_ERROR);
	else if(current_ua < 400)  // About 2mW at 5V
		LP5861_led_setState(RGB_LED_READY);
	else
		LP5861_led_setState(RGB_LED_RUNNING);

    tlc59116_setDriverOnLed(pumpOn);
    uint16_t power_mw = ( (sampling_getChannelFiltered(DRIVE_MICRO_AMPS)/10 ) *
            (sampling_getChannelFiltered(DRIVE_MILLI_VOLTS)/10) )
          / 10000;
    tlc59116_setGaugemW(power_mw);
    LP5861_setGaugemW(power_mw);

    LP5861_setLEDColour((uint16_t) settings_getHandle()->LED_Colour->value);
    LP5861_tick();
}

